from __future__ import annotations

import os
import json
import typing as t
import platform
import tempfile
import threading
from http import HTTPStatus
from types import TracebackType
from pathlib import Path
from functools import wraps, partial, lru_cache
from itertools import islice

import yaml
from loguru import logger

from starwhale.utils import now_str, convert_to_bytes, gen_uniq_version
from starwhale.consts import (
    FileDesc,
    HTTPMethod,
    CREATED_AT_KEY,
    SW_TMP_DIR_NAME,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    STANDALONE_INSTANCE,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.version import STARWHALE_VERSION
from starwhale.base.uri import URI, URIType
from starwhale.utils.fs import copy_file, empty_dir, ensure_dir, ensure_file
from starwhale.base.type import InstanceType
from starwhale.base.cloud import CloudRequestMixed
from starwhale.utils.error import NoSupportError
from starwhale.utils.config import SWCliConfigMixed
from starwhale.core.dataset.type import (
    DatasetSummary,
    D_ALIGNMENT_SIZE,
    D_FILE_VOLUME_SIZE,
)
from starwhale.core.dataset.model import Dataset as CoreDataset
from starwhale.core.dataset.model import StandaloneDataset
from starwhale.core.dataset.store import DatasetStorage
from starwhale.api._impl.data_store import TableEmptyException
from starwhale.core.dataset.tabular import (
    TabularDataset,
    DatastoreRevision,
    TabularDatasetInfo,
    DatastoreWrapperDataset,
    get_dataset_consumption,
    DEFAULT_CONSUMPTION_BATCH_SIZE,
    TabularDatasetSessionConsumption,
)

from .loader import DataRow, DataLoader, get_data_loader
from .builder import MappingDatasetBuilder

if t.TYPE_CHECKING:
    import tensorflow as tf
    from torch.utils.data import Dataset as TorchDataset

_DType = t.TypeVar("_DType", bound="Dataset")
_ItemType = t.Union[str, int, slice]
_GItemType = t.Optional[t.Union[DataRow, t.List[DataRow]]]

_DEFAULT_LOADER_WORKERS = 2
_DEFAULT_LOADER_CACHE_SIZE = 20


class _DatasetCreateMode:
    auto = "auto"
    empty = "empty"
    forbid = "forbid"


class _Tags:
    def __init__(self, core_dataset: CoreDataset) -> None:
        self.__core_dataset = core_dataset

    def add(self, tags: t.Union[str, t.List[str]], ignore_errors: bool = False) -> None:
        if isinstance(tags, str):
            tags = [tags]
        self.__core_dataset.add_tags(tags, ignore_errors)

    def remove(
        self, tags: t.Union[str, t.List[str]], ignore_errors: bool = False
    ) -> None:
        if isinstance(tags, str):
            tags = [tags]
        self.__core_dataset.remove_tags(tags, ignore_errors)

    def __iter__(self) -> t.Generator[str, None, None]:
        for tag in self.__core_dataset.list_tags():
            yield tag

    def __str__(self) -> str:
        return f"Dataset Tag: {self.__core_dataset}"

    __repr__ = __str__


class Dataset:
    def __init__(
        self,
        name: str,
        version: str,
        project_uri: URI,
        readonly: bool = False,
        create: str = _DatasetCreateMode.auto,
    ) -> None:
        self.name = name
        self.project_uri = project_uri
        self.__readonly = readonly
        self._pending_commit_version = gen_uniq_version()

        self._make_capsulated_uri = partial(
            URI.capsulate_uri,
            instance=self.project_uri.instance,
            project=self.project_uri.project,
            obj_type=URIType.DATASET,
            obj_name=self.name,
        )

        if create not in (
            _DatasetCreateMode.auto,
            _DatasetCreateMode.empty,
            _DatasetCreateMode.forbid,
        ):
            raise ValueError(
                f"the current create mode is not in the accept options: {create}"
            )

        _origin_uri = self._make_capsulated_uri(obj_ver=version or "latest")
        origin_uri_exists = self._check_uri_exists(_origin_uri)
        if origin_uri_exists:
            if create == _DatasetCreateMode.empty:
                raise RuntimeError(
                    f"dataset already existed, failed to create by the {create} create mode: {self.name}"
                )

            self._loading_version = self._auto_complete_version(
                _origin_uri.object.version
            )
        else:
            # TODO: call server or standalone api to create an empty dataset before committing.
            if create == _DatasetCreateMode.forbid:
                raise RuntimeError(
                    "dataset doest not exist, we have already use {create} create mode to ensure dataset existence: {self.name}"
                )

            if readonly:
                raise ValueError(
                    f"no support to set a non-existed dataset to the readonly mode: {self.name}"
                )

            if version != "":
                raise NoSupportError(
                    f"no support to create a specified version dataset: {version}"
                )

            self._loading_version = self._pending_commit_version

        self._loading_uri = self._make_capsulated_uri(obj_ver=self._loading_version)
        self.__loading_core_dataset = CoreDataset.get_dataset(self._loading_uri)

        self._pending_commit_uri = self._make_capsulated_uri(
            obj_ver=self._pending_commit_version
        )
        self.__pending_commit_core_dataset = CoreDataset.get_dataset(
            self._pending_commit_uri
        )

        self._consumption: t.Optional[TabularDatasetSessionConsumption] = None
        self._loader_lock = threading.Lock()
        self.__data_loaders: t.Dict[str, DataLoader] = {}
        self._loader_cache_size = _DEFAULT_LOADER_CACHE_SIZE
        self._loader_num_workers = _DEFAULT_LOADER_WORKERS

        self._builder_blob_alignment_size = D_ALIGNMENT_SIZE
        self._builder_blob_volume_size = D_FILE_VOLUME_SIZE
        self._builder_lock = threading.Lock()
        self._dataset_builder: t.Optional[MappingDatasetBuilder] = None
        self._commit_lock = threading.Lock()
        self.__has_committed = False

        self._info_lock = threading.Lock()
        self._info_ds_wrapper: t.Optional[DatastoreWrapperDataset] = None
        self.__info: t.Optional[TabularDatasetInfo] = None

        self._tmpdir: t.Optional[Path] = None

        self._updated_rows_by_commit = 0
        self._deleted_rows_by_commit = 0
        if origin_uri_exists:
            _summary = self.__loading_core_dataset.summary()
            # TODO: raise none summary exception for existed dataset
            if _summary is None:
                self._total_rows = 0
                self._total_blobs_size = 0
            else:
                self._total_rows = _summary.rows
                self._total_blobs_size = _summary.blobs_byte_size
        else:
            self._total_rows = 0
            self._total_blobs_size = 0

        self._last_data_datastore_revision = ""
        self._last_info_datastore_revision = ""

    def _auto_complete_version(self, version: str) -> str:
        version = version.strip()
        if not version:
            return version

        # TODO: auto complete for cloud instance
        if self.project_uri.instance_type == InstanceType.CLOUD:
            return version

        _uri = self._make_capsulated_uri(obj_ver=version)
        store = DatasetStorage(_uri)
        if not store.snapshot_workdir.exists():
            return version
        else:
            return store.id

    def __str__(self) -> str:
        return f"Dataset: {self.name}, stash version: {self._pending_commit_version}, loading version: {self._loading_version}"

    def __repr__(self) -> str:
        return f"Dataset: {self.name}, loading uri: {self._loading_uri}, pending commit uri: {self._pending_commit_uri}"

    def __len__(self) -> int:
        return self._total_rows

    def __enter__(self: _DType) -> _DType:
        return self

    def __bool__(self) -> bool:
        return True

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        trace: TracebackType,
    ) -> None:
        if value:  # pragma: no cover
            logger.warning(f"type:{type}, exception:{value}, traceback:{trace}")

        self.close()

    def make_distributed_consumption(
        self, session_id: str, batch_size: int = DEFAULT_CONSUMPTION_BATCH_SIZE
    ) -> Dataset:
        if self._consumption is not None:
            raise RuntimeError(
                f"distributed consumption has already been created ({self._consumption})"
            )

        with self._loader_lock:
            self._consumption = get_dataset_consumption(
                self.uri, session_id=session_id, batch_size=batch_size
            )
        return self

    def _clear_data_loader(self) -> None:
        with self._loader_lock:
            self.__data_loaders = {}

    def with_loader_config(
        self, num_workers: t.Optional[int] = None, cache_size: t.Optional[int] = None
    ) -> Dataset:
        with self._loader_lock:
            if len(self.__data_loaders) != 0:
                raise RuntimeError(
                    f"loaders({list(self.__data_loaders)}) have already been initialized"
                )

            if num_workers is not None:
                self._loader_num_workers = num_workers

            if cache_size is not None:
                self._loader_cache_size = cache_size

        return self

    def with_builder_blob_config(
        self,
        volume_size: int | str = D_FILE_VOLUME_SIZE,
        alignment_size: int | str = D_ALIGNMENT_SIZE,
    ) -> Dataset:
        """Config blob attributes for the dataset builder.

        If you want to config blob attributes, you should call the function before appending, updating or deleting dataset records.

        Arguments:
            volume_size: (str, int, optional) The max bytes size of the single blob file.
                When blob file size exceeds the value of volume_size argument, a new blob file is created automatically.
                The default is 64MB. The argument accepts int(bytes) or str(32K, 64MB, 1GB...) format.
            alignment_size: (str, int, optional) The alignment size of every bin section in the blob file.
                The remaining part will be filled with `\0`. Default is 64 bytes.

        Examples:
        ```python
        from starwhale import dataset, Binary

        ds = dataset("mnist").with_builder_blob_config(volume_size="32M", alignment_size=128)
        ds.append({"data": Binary(b"123")})
        ds.commit()
        ds.close()
        ```

        Returns:
            A Dataset Object
        """
        with self._builder_lock:
            if self._dataset_builder is not None:
                raise RuntimeError(
                    "dataset has already accept some changed rows, forbid to config dataset blob attributes"
                )

            self._builder_blob_volume_size = convert_to_bytes(volume_size)
            self._builder_blob_alignment_size = convert_to_bytes(alignment_size)

        return self

    def _get_data_loader(
        self, recreate: bool = False, disable_consumption: bool = False
    ) -> DataLoader:
        with self._loader_lock:
            key = f"consumption-{disable_consumption}-{self.uri.object.version}"
            _loader = self.__data_loaders.get(key)
            if _loader is None or recreate:
                if disable_consumption:
                    consumption = None
                else:
                    consumption = self._consumption

                data_revision = self._last_data_datastore_revision
                if data_revision == "":
                    data_revision = self._get_datastore_revision(self.uri).data

                _loader = get_data_loader(
                    self.uri,
                    session_consumption=consumption,
                    cache_size=self._loader_cache_size,
                    num_workers=self._loader_num_workers,
                    dataset_scan_revision=data_revision,
                )
                self.__data_loaders[key] = _loader

        return _loader

    @lru_cache(maxsize=32)
    def _get_datastore_revision(self, uri: URI) -> DatastoreRevision:
        if uri.object.typ != URIType.DATASET:
            raise NoSupportError(
                f"only support to fetch dataset datastore revision: {uri}"
            )

        if uri.object.version == "":
            raise RuntimeError(f"cannot get version of uri: {uri}")

        if uri.instance_type == InstanceType.CLOUD:
            crm = CloudRequestMixed()
            r = crm.do_http_request(
                path=f"/project/{uri.project}/dataset/{uri.object.name}",
                instance_uri=uri,
                params={"versionUrl": uri.object.version},
            ).json()
            manifest = yaml.safe_load(r["data"]["versionMeta"])
        else:
            manifest = DatasetStorage(uri).manifest

        return DatastoreRevision.from_manifest(manifest)

    def batch_iter(
        self, batch_size: int = 1, drop_not_full: bool = False
    ) -> t.Iterator[t.List[DataRow]]:
        """Batch data into lists of length n. The last batch may be shorter."""
        it = self.__iter__()
        while True:
            batch_data = list(islice(it, batch_size))
            if not batch_data or (drop_not_full and len(batch_data) < batch_size):
                return
            yield batch_data

    def __iter__(self) -> t.Iterator[DataRow]:
        for row in self._get_data_loader():
            row._patch_shadow_dataset(self)
            yield row

    def __getitem__(
        self,
        item: _ItemType,
    ) -> _GItemType:
        """
        Example:
            self["str_key"]  # get the DataRow by the "str_key" string key
            self[1]  # get the DataRow by the 1 int key
            self["start":"end"]  # get a slice of the dataset by the range ("start", "end")
            self[1:10:2] # get a slice of the dataset by the range (1, 10), step is 2
        """
        # TODO: tune datastore performance for getitem
        return self._getitem(item)

    def _getitem(
        self,
        item: _ItemType,
        skip_fetch_data: bool = False,
    ) -> _GItemType:
        def _run() -> _GItemType:
            loader = self._get_data_loader(disable_consumption=True)
            if isinstance(item, (int, str)):
                row = next(loader.tabular_dataset.scan(item, item, end_inclusive=True))
                return loader._unpack_row(row, skip_fetch_data, shadow_dataset=self)
            elif isinstance(item, slice):
                step = item.step or 1
                if step <= 0:
                    raise ValueError(
                        f"Dataset slice step({step}) cannot be zero or negative number"
                    )
                cnt = 0
                # TODO: batch signed urls
                rows = []
                for row in loader.tabular_dataset.scan(item.start, item.stop):
                    if cnt % step == 0:
                        rows.append(
                            loader._unpack_row(
                                row,
                                skip_fetch_data,
                                shadow_dataset=self,
                            )
                        )
                    cnt += 1
                return rows
            else:
                raise ValueError(f"{item} type is not int, str or slice")

        try:
            return _run()
        except TableEmptyException:
            self._clear_data_loader()
            return None
        except StopIteration:
            return None

    @property
    def readonly(self) -> bool:
        return self.__readonly

    def _check_readonly(func: t.Callable):  # type: ignore
        @wraps(func)
        def _wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
            self: Dataset = args[0]
            if self.readonly:
                raise RuntimeError(f"{func} does not work in the readonly mode")
            return func(*args, **kwargs)

        return _wrapper

    @property
    def tags(self) -> _Tags:
        return _Tags(self.__loading_core_dataset)

    @staticmethod
    def _check_uri_exists(uri: t.Optional[URI]) -> bool:
        if uri is None or uri.object.version == "":
            return False

        if uri.instance_type == InstanceType.CLOUD:
            crm = CloudRequestMixed()
            ok, _ = crm.do_http_request_simple_ret(
                path=f"/project/{uri.project}/{URIType.DATASET}/{uri.object.name}/version/{uri.object.version}",
                method=HTTPMethod.HEAD,
                instance_uri=uri,
                ignore_status_codes=[HTTPStatus.NOT_FOUND],
            )
            return ok
        else:
            _store = DatasetStorage(uri)
            return _store.manifest_path.exists()

    def exists(self) -> bool:
        return self._check_uri_exists(self.uri)

    @property
    def info(self) -> TabularDatasetInfo:
        with self._info_lock:
            if self.__info is None:
                info_revision = self._last_info_datastore_revision
                if info_revision == "":
                    info_revision = self._get_datastore_revision(self.uri).info

                self._info_ds_wrapper = TabularDataset.from_uri(
                    self.uri, info_datastore_revision=info_revision
                )._info_ds_wrapper
                self.__info = TabularDatasetInfo.load_from_datastore(
                    self._info_ds_wrapper
                )

            return self.__info

    @_check_readonly
    def flush(self, artifacts_flush: bool = False) -> str:
        revision = ""
        if self._dataset_builder:
            revision = self._dataset_builder.flush(artifacts_flush)
            self._last_data_datastore_revision = revision
            self._clear_data_loader()

        return revision

    @_check_readonly
    def rehash(self) -> None:
        # TODO: rehash for swds-bin format dataset with append/delete items to reduce volumes size
        raise NotImplementedError

    def remove(self, force: bool = False) -> None:
        ok, reason = self.__loading_core_dataset.remove(force)
        if not ok:
            raise RuntimeError(f"failed to remove dataset: {reason}")

    def recover(self, force: bool = False) -> None:
        ok, reason = self.__loading_core_dataset.recover(force)
        if not ok:
            raise RuntimeError(f"failed to recover dataset: {reason}")

    def summary(self) -> t.Optional[DatasetSummary]:
        return self.__loading_core_dataset.summary()

    def history(self) -> t.List[t.Dict]:
        return self.__loading_core_dataset.history()

    def close(self) -> None:
        for _loader in self.__data_loaders.values():
            if not _loader:
                continue  # pragma: no cover
            _loader.tabular_dataset.close()

        if self._dataset_builder:
            self._dataset_builder.close()

        if self._tmpdir is not None:
            empty_dir(self._tmpdir)

    @property
    def tmpdir(self) -> Path:
        if self._tmpdir is None:
            _base_dir = SWCliConfigMixed().rootdir / SW_TMP_DIR_NAME
            ensure_dir(_base_dir)
            self._tmpdir = Path(
                tempfile.mkdtemp(prefix=f"dataset-{self.name}-", dir=_base_dir)
            )
        return self._tmpdir

    def diff(self, cmp: Dataset) -> t.Dict:
        # TODO: wait for datastore diff feature
        raise NotImplementedError

    def manifest(self) -> t.Dict[str, t.Any]:
        return self.__loading_core_dataset.info()

    def head(self, n: int = 5, skip_fetch_data: bool = False) -> t.List[DataRow]:
        # TODO: render artifact in JupyterNotebook
        ret = []
        loader = self._get_data_loader(disable_consumption=True)
        for idx, td_row in enumerate(loader._iter_meta()):
            if idx >= n:
                break
            data_row = loader._unpack_row(td_row, skip_fetch_data=skip_fetch_data)
            ret.append(data_row)
        return ret

    def fetch_one(self, skip_fetch_data: bool = False) -> DataRow:
        return self.head(1, skip_fetch_data)[0]

    def to_pytorch(
        self,
        transform: t.Optional[t.Callable] = None,
        drop_index: bool = True,
        skip_default_transform: bool = False,
    ) -> TorchDataset:
        from starwhale.integrations.pytorch import TorchIterableDataset

        return TorchIterableDataset(
            dataset=self,
            transform=transform,
            drop_index=drop_index,
            skip_default_transform=skip_default_transform,
        )

    def to_tensorflow(
        self,
        drop_index: bool = True,
    ) -> tf.data.Dataset:
        from starwhale.integrations.tensorflow import to_tf_dataset

        return to_tf_dataset(dataset=self, drop_index=drop_index)

    @_check_readonly
    def __setitem__(
        self, key: t.Union[str, int], value: t.Union[DataRow, t.Tuple, t.Dict]
    ) -> None:
        # TODO: tune the performance of getitem by cache
        if not isinstance(key, (int, str)):
            raise TypeError(f"key must be str or int type: {key}")

        if isinstance(value, DataRow):
            value.index = key
            row = value
        elif isinstance(value, dict):
            row = DataRow(index=key, features=value)
        elif isinstance(value, (tuple, list)):
            if len(value) == 1:
                features = value[0]
            elif len(value) == 2:
                _, features = value
            else:
                raise ValueError(f"{value} cannot unpack")

            row = DataRow(index=key, features=features)
        else:
            raise TypeError(f"value only supports tuple, dict or DataRow type: {value}")

        # TODO: add gc/rehash for update swds-bin format artifact features
        # TODO improve accuracy of _total_rows during building
        self._total_rows += 1

        _ds_builder = self._get_dataset_builder()
        _ds_builder.put(row)
        self._updated_rows_by_commit += 1

    def _get_dataset_builder(self) -> MappingDatasetBuilder:
        with self._builder_lock:
            if self._dataset_builder is None:
                # TODO: support alignment_bytes_size, volume_bytes_size arguments
                self._dataset_builder = MappingDatasetBuilder(
                    workdir=self.tmpdir / "builder",
                    dataset_name=self.name,
                    project_name=self.project_uri.project,
                    instance_name=self.project_uri.instance,
                    blob_alignment_bytes_size=self._builder_blob_alignment_size,
                    blob_volume_bytes_size=self._builder_blob_volume_size,
                )
            return self._dataset_builder

    @_check_readonly
    def __delitem__(self, key: _ItemType) -> None:
        items: t.List
        if isinstance(key, (str, int)):
            items = [self._getitem(key, skip_fetch_data=True)]
        elif isinstance(key, slice):
            items = self._getitem(key, skip_fetch_data=True)  # type: ignore
        else:
            raise TypeError(f"key({key}) is not str, int or slice type")

        # TODO: raise not-found key error?
        _ds_builder = self._get_dataset_builder()
        for item in items:
            if not item or not isinstance(item, DataRow):
                continue  # pragma: no cover
            _ds_builder.delete(item.index)
            self._deleted_rows_by_commit += 1
            self._total_rows -= 1

    @_check_readonly
    def append(self, item: t.Any) -> None:
        if isinstance(item, DataRow):
            self.__setitem__(item.index, item)
        elif isinstance(item, dict):
            self.__setitem__(self._total_rows, item)
        elif isinstance(item, (list, tuple)):
            if len(item) == 1:
                row = DataRow(self._total_rows, item[0])
            elif len(item) == 2:
                row = DataRow(item[0], item[1])
            else:
                raise ValueError(
                    f"cannot unpack value({item}), expected sequence is (index, data) or data"
                )

            self.__setitem__(row.index, row)
        else:
            raise TypeError(f"value({item}) is not DataRow, list or tuple type")

    @_check_readonly
    def extend(self, items: t.Sequence[t.Any]) -> None:
        for item in items:
            self.append(item)

    @property
    def loading_version(self) -> str:
        """Loaded dataset version.
        When loading an existing dataset, the loading_version is the related dataset version.
        When creating a non-existed dataset, the loading_version is equal to the pending_commit_version.

        Returns:
            A str version
        """
        return self._loading_version

    @property
    def pending_commit_version(self) -> str:
        """Next commit version.
        When you call the `commit` function, the pending_commit_version will be recorded in
        the Standalone instance ,or Cloud instance.

        Returns:
            A str version
        """
        return self._pending_commit_version

    @property
    def committed_version(self) -> str:
        """Committed version.

        After the commit function is called, the committed_version will come out, it is equal to the pending_commit_version.

        Returns:
            A str version
        """
        if self.__has_committed:
            return self._pending_commit_version
        else:
            raise RuntimeError("version has not been committed yet")

    @property
    def uri(self) -> URI:
        """Dataset URI

        Before committing, the uri is for the loading_version.
        After committing, the uri is for the committed_version.

        Returns:
            A URI object.
        """
        if self.__has_committed:
            return self._pending_commit_uri
        else:
            return self._loading_uri

    @property
    def changed(self) -> bool:
        """For the non-readonly dataset, when the users update/delete/add records(rows) or features(columns),
        changed property will change from False to True.

        Returns:
            True or False
        """
        return self._updated_rows_by_commit != 0 or self._deleted_rows_by_commit != 0

    @_check_readonly
    def commit(self, tags: t.Optional[t.List[str]] = None, message: str = "") -> str:
        """Commit into dataset
        Commit will flush and generate a version of the dataset. At the same time, commit
        operation will also generate auto-increment tag, such as v0, v1, v2. Only one commit is allowed.

        Arguments:
            tags: (list(str), optional) Specify the tags for the version. Default is None.
            message: (str, optional) Commit message. Default is empty str.

        Example:
        ```python
        from starwhale import dataset
        with dataset("mnist") as ds:
            ds.append({"label": 1})
            ds.commit(message="init commit")
        ```

        Return:
            A str of dataset version.
        """
        # TODO: forbid commit many times
        with self._commit_lock:
            return self._commit(tags or [], message)

    def _commit(self, tags: t.List[str], message: str) -> str:
        def _save_info() -> str:
            revision = ""
            if self.__info is not None and self._info_ds_wrapper is not None:
                revision = self.__info.save_to_datastore(self._info_ds_wrapper)
                self._last_info_datastore_revision = revision
            return revision

        def _dump_manifest(dataset_revision: str, info_revision: str) -> Path:
            if self._dataset_builder is None:
                raise RuntimeError("failed to commit, because dataset builder is None")

            increased_blobs_size = sum(
                [m.size for m in self._dataset_builder.signature_bins_meta]
            )

            _manifest = {
                "build": {
                    "os": platform.system(),
                    "starwhale": STARWHALE_VERSION,
                },
                "version": self._pending_commit_version,
                CREATED_AT_KEY: now_str(),
                "dataset_summary": DatasetSummary(
                    rows=self._dataset_builder.calculate_rows_cnt(),  # maybe slow
                    updated_rows=self._updated_rows_by_commit,
                    deleted_rows=self._deleted_rows_by_commit,
                    blobs_byte_size=self._total_blobs_size + increased_blobs_size,
                    increased_blobs_byte_size=increased_blobs_size,
                ).asdict(),
                "message": message,
            }

            _manifest.update(
                DatastoreRevision(data=dataset_revision, info=info_revision).asdict()
            )

            self._updated_rows_by_commit = 0
            self._deleted_rows_by_commit = 0

            _m_path = self.tmpdir / DEFAULT_MANIFEST_NAME
            ensure_file(
                _m_path,
                yaml.safe_dump(_manifest, default_flow_style=False),
                parents=True,
            )
            return _m_path

        def _submit_standalone_version(manifest_path: Path) -> None:
            pccd = self.__pending_commit_core_dataset
            if not isinstance(pccd, StandaloneDataset):
                raise RuntimeError(
                    f"core dataset is not StandaloneDataset instance: {pccd}"
                )

            _snapshot_dir = pccd.store.snapshot_workdir
            ensure_dir(_snapshot_dir)
            copy_file(
                manifest_path,
                pccd.store.snapshot_workdir / DEFAULT_MANIFEST_NAME,
            )
            pccd.tag.add_fast_tag()

        def _submit_cloud_version(manifest_path: Path) -> None:
            from starwhale.base.bundle_copy import _UploadPhase

            crm = CloudRequestMixed()
            params = {
                "swds": f"{self.name}:{self._pending_commit_version}",
                "project": self.project_uri.project,
                "force": "1",  # stash version is unique, use force=1 to make http retry happy
            }
            url_path = f"/project/{self.project_uri.project}/dataset/{self.name}/version/{self._pending_commit_version}/file"
            r = crm.do_multipart_upload_file(
                url_path=url_path,
                file_path=manifest_path,
                instance_uri=self.project_uri,
                params={
                    "phase": _UploadPhase.MANIFEST,
                    "desc": FileDesc.MANIFEST.name,
                    **params,
                },
                use_raise=True,
            )
            crm.do_http_request(
                path=url_path,
                method=HTTPMethod.POST,
                instance_uri=self.project_uri,
                data={
                    "phase": _UploadPhase.END,
                    "uploadId": r.json()["data"]["uploadId"],
                    **params,
                },
                use_raise=True,
                disable_default_content_type=True,
            )

        if self.__has_committed:
            raise RuntimeError("Dataset has already committed")

        dataset_revision = self.flush(artifacts_flush=True)
        info_revision = _save_info()
        manifest_path = _dump_manifest(dataset_revision, info_revision)

        logger.debug(
            f"dataset commit: revision-{dataset_revision}, info revision-{info_revision}"
        )
        if self.project_uri.instance == STANDALONE_INSTANCE:
            _submit_standalone_version(manifest_path)
        else:
            _submit_cloud_version(manifest_path)
        os.unlink(manifest_path)

        if tags:
            self.__pending_commit_core_dataset.add_tags(tags)

        self.__has_committed = True
        return self._pending_commit_version

    @property
    def committed(self) -> bool:
        """Check If the dataset is committed.

        Returns:
            True or False
        """
        return self.__has_committed

    def copy(self, dest_uri: str, dest_local_project_uri: str = "") -> None:
        CoreDataset.copy(
            str(self.uri),
            dest_uri,
            dest_local_project_uri=dest_local_project_uri,
        )

    @classmethod
    def list(
        cls,
        project_uri: t.Union[str, URI] = "",
        fullname: bool = False,
        show_removed: bool = False,
        page_index: int = DEFAULT_PAGE_IDX,
        page_size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        from starwhale.core.dataset.view import DatasetTermView

        return DatasetTermView.list(
            project_uri, fullname, show_removed, page_index, page_size
        )

    @classmethod
    def dataset(
        cls,
        uri: t.Union[str, URI],
        create: str = _DatasetCreateMode.auto,
        readonly: bool = False,
    ) -> Dataset:
        """Create or load a dataset from standalone instance or cloud instance.

        For a non-existed dataset, when you try to load it, you will get an Exception; when you try to append records and commit,
        the dataset will be created automatically.

        Arguments:
            uri: (str, URI, required) The dataset uri.
            create: (str, optional) The mode of dataset creating. The options are `auto`, `empty` and `forbid`.
                `auto` mode: If the dataset already exists, creation is ignored. If it does not exist, the dataset is created automatically.
                `empty` mode: If the dataset already exists, an Exception is raised; If it does not exist, an empty dataset is created. This mode ensures the creation of a new, empty dataset.
                `forbid` mode: If the dataset already exists, nothing is done.If it does not exist, an Exception is raised. This mode ensures the existence of the dataset.
                The default is `auto`.
            readonly: (bool, optional) For an existing dataset, you can specify the readonly=True argument to ensure
                the dataset is in readonly mode. Default is False.

        Returns:
            A Dataset Object

        Examples:
        ```python
        from starwhale import dataset, Image

        # create a new dataset named mnist, and add a row into the dataset
        # dataset("mnist") is equal to dataset("mnist", create="auto")
        ds = dataset("mnist")
        ds.exists()  # return False, "mnist" dataset is not existing.
        ds.append({"img": Image(), "label": 1})
        ds.commit()
        ds.close()

        # load a cloud instance dataset in readonly mode
        ds = dataset("cloud://remote-instance/project/starwhale/dataset/mnist", readonly=True)
        labels = [row.features.label in ds]
        ds.close()

        # load a read/write dataset with a specified version
        ds = dataset("mnist/version/mrrdczdbmzsw")
        ds[0].features.label = 1
        ds.commit()
        ds.close()

        # create an empty dataset
        ds = dataset("mnist-empty", create="empty")

        # ensure the dataset existence
        ds = dataset("mnist-existed", create="forbid")
        ```

        """
        if isinstance(uri, str):
            _uri = URI(uri, expected_type=URIType.DATASET)
        elif isinstance(uri, URI) and uri.object.typ == URIType.DATASET:
            _uri = uri
        else:
            raise TypeError(
                f"uri({uri}) argument type is not expected, dataset uri or str is ok"
            )

        ds = Dataset(
            name=_uri.object.name,
            version=_uri.object.version,
            project_uri=_uri,  # TODO: cut off dataset resource info?
            readonly=readonly,
            create=create or _DatasetCreateMode.auto,
        )

        return ds

    @classmethod
    def from_json(cls, name: str, json_text: str, field_selector: str = "") -> Dataset:
        """Create a new dataset from a dict.
        Arguments:
            name: (str, required) The dataset name you would like to use.
            json_text: (str, required) The json text from which you would like to create this dataset
            field_selector: (str, optional) The filed from which you would like to extract dataset array items.
                The default value is "" which indicates that the dict is an array contains all the items.
            Returns:
                A Dataset Object
            Examples:
            ```python
            from starwhale import Dataset
            myds = Dataset.from_json("translation", '[{"en":"hello","zh-cn":"你好"},{"en":"how are you","zh-cn":"最近怎么样"}]')
            print(myds[0].features.en)
            ```
            ```python
            from starwhale import Dataset
            myds = Dataset.from_json("translation", '{"content":{"child_content":[{"en":"hello","zh-cn":"你好"},{"en":"how are you","zh-cn":"最近怎么样"}]}}',"content.child_content")
            print(myds[0].features["zh-cn"])
            ```
        """
        data_items = json.loads(json_text)
        if field_selector:
            # Split field selector by dots
            fields = field_selector.split(".")
            # Iterate over selected fields
            for field in fields:
                if field in data_items:
                    data_items = data_items[field]
                else:
                    raise ValueError(
                        f"The field_selector {field_selector} isn't in json_text: {json_text}"
                    )
        if not isinstance(data_items, list):
            raise ValueError(
                f"The field selected by field_selector {field_selector} isn't an array: {data_items}"
            )
        ds = cls.dataset(name)
        for item in data_items:
            ds.append(item)
        ds.commit()
        ds.close()
        return ds



import json
import base64
import typing as t
from copy import deepcopy
from enum import Enum

from starwhale.utils.error import FormatError


class ASDictMixin:
    def asdict(self, ignore_keys: t.Optional[t.List[str]] = None) -> t.Dict:
        d = deepcopy(self.__dict__)
        ignore_keys = ignore_keys or []
        for _k in ignore_keys:
            d.pop(_k, None)

        r = _do_asdict_convert(d)
        if isinstance(r, dict):
            return r
        else:
            raise FormatError(f"{self} cannot be formatted as a dict")

    def jsonify(self, ignore_keys: t.Optional[t.List[str]] = None) -> str:
        r = self.asdict(ignore_keys)
        return json.dumps(r, separators=(",", ":"))


def _do_asdict_convert(obj: t.Any) -> t.Any:
    if isinstance(obj, dict):
        return {k: _do_asdict_convert(v) for k, v in obj.items()}
    elif (
        hasattr(obj, "_asdict") and hasattr(obj, "_fields") and isinstance(obj, tuple)
    ):  # namedtuple
        return {k: _do_asdict_convert(v) for k, v in obj._asdict().items()}  # type: ignore
    elif isinstance(obj, (list, tuple)):
        return type(obj)(_do_asdict_convert(v) for v in obj)
    elif isinstance(obj, Enum):
        return obj.value
    elif hasattr(obj, "asdict"):
        return obj.asdict()
    elif isinstance(obj, bytes):
        return base64.b64encode(obj).decode()
    else:
        # TODO: add more type parse
        return obj



import os
import typing as t
from subprocess import PIPE, Popen, STDOUT, CalledProcessError

from loguru import logger


def log_check_call(*args: t.Any, **kwargs: t.Any) -> int:
    log = kwargs.pop("log", logger.debug)
    kwargs["bufsize"] = 1
    kwargs["stdout"] = PIPE
    kwargs["stderr"] = STDOUT
    env = os.environ.copy()
    env.update(kwargs.get("env", {}))
    env["PYTHONUNBUFFERED"] = "1"
    kwargs["env"] = env
    kwargs["universal_newlines"] = True

    output = []
    p = Popen(*args, **kwargs)
    logger.debug(f"cmd: {p.args!r}")

    while True:
        line = p.stdout.readline()  # type: ignore
        if line:
            log(line.rstrip())
            output.append(line)

        if p.poll() is not None:
            break

    p.wait()
    for line in p.stdout.readlines():  # type: ignore
        if line:
            log(line.rstrip())
            output.append(line)

    try:
        p.stdout.close()  # type: ignore
    except Exception as ex:
        log(f"failed to close stdout:{ex}")

    if p.returncode != 0:
        cmd = kwargs.get("args") or args[0]
        e = CalledProcessError(p.returncode, cmd)
        e.output = "".join(output)
        raise e

    return 0


check_call = log_check_call



import os
import shutil
import typing as t
import tarfile
import tempfile
from pathlib import Path

from starwhale.utils.error import FormatError, ExistedError


def pack(prefix: t.Union[str, Path], output: str, force: bool = False) -> None:
    if os.path.exists(output) and not force:
        raise ExistedError(output)

    from .venv import check_valid_venv_prefix

    if not check_valid_venv_prefix(prefix):
        raise FormatError(f"venv prefix: {prefix}")

    _, temp_fpath = tempfile.mkstemp(
        prefix="starwhale-venv-", dir=os.path.dirname(output)
    )

    # TODO: check editable packages
    try:
        with tarfile.open(temp_fpath, "w:gz") as tar:
            tar.add(str(prefix), arcname="")

        shutil.move(temp_fpath, output)
    finally:
        if os.path.exists(temp_fpath):
            os.unlink(temp_fpath)



import typing as t
from copy import deepcopy
from http import HTTPStatus
from pathlib import Path
from datetime import datetime, timedelta

import yaml
import requests
from rich.progress import TaskID, Progress
from requests_toolbelt.multipart.encoder import (  # type: ignore
    MultipartEncoder,
    MultipartEncoderMonitor,
)

from starwhale.consts import (
    HTTPMethod,
    FMT_DATETIME,
    CREATED_AT_KEY,
    SW_API_VERSION,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
)
from starwhale.base.uri import URI
from starwhale.utils.fs import ensure_dir
from starwhale.utils.http import ignore_error, wrap_sw_error_resp
from starwhale.utils.error import NoSupportError
from starwhale.utils.retry import http_retry

_TMP_FILE_BUFSIZE = 8192
_DEFAULT_TIMEOUT_SECS = 90
_UPLOAD_CHUNK_SIZE = 20 * 1024 * 1024


class CloudRequestMixed:
    def fmt_timestamp(self, ts: t.Union[float, str]) -> str:
        return datetime.fromtimestamp(float(ts) / 1000.0).strftime(FMT_DATETIME).strip()

    def fmt_duration(self, ts: t.Union[float, str]) -> str:
        return str(timedelta(milliseconds=float(ts)))

    def do_download_file(
        self,
        url_path: str,
        dest_path: Path,
        instance_uri: URI,
        progress: t.Optional[Progress] = None,
        task_id: TaskID = TaskID(0),
        **kw: t.Any,
    ) -> None:
        r = self.do_http_request(
            path=url_path,
            instance_uri=instance_uri,
            method=HTTPMethod.GET,
            use_raise=True,
            **kw,
        )
        total = float(r.headers.get("content-length", 0))
        ensure_dir(dest_path.parent)
        with dest_path.open("wb") as f:
            for chunk in r.iter_content(chunk_size=_TMP_FILE_BUFSIZE):
                if progress:
                    progress.update(task_id, total=total, advance=len(chunk))
                f.write(chunk)
        if progress:
            # TODO: remove the hack code when api support content-length header
            progress.update(task_id, total=1, completed=1)

    def do_multipart_upload_file(
        self,
        url_path: str,
        instance_uri: URI,
        file_path: t.Union[str, Path],
        fields: t.Optional[t.Dict[str, t.Any]] = None,
        headers: t.Optional[t.Dict[str, t.Any]] = None,
        progress: t.Optional[Progress] = None,
        task_id: TaskID = TaskID(0),
        **kw: t.Any,
    ) -> requests.Response:
        fields = fields or {}
        headers = headers or {}
        # TODO: add progress bar and rich live

        def _progress_bar(monitor: MultipartEncoderMonitor) -> None:
            if progress:
                progress.update(task_id, completed=monitor.bytes_read)

        _headers = deepcopy(headers)
        fpath = Path(file_path).resolve().absolute()

        with fpath.open("rb") as f:
            fields["file"] = (fpath.name, f, "text/plain")

            _encoder = MultipartEncoder(fields=fields)
            # default chunk is 8192 Bytes
            _encoder._read = _encoder.read  # type: ignore
            _encoder.read = lambda size: _encoder._read(_UPLOAD_CHUNK_SIZE)  # type: ignore
            _headers["Content-Type"] = _encoder.content_type
            _monitor = MultipartEncoderMonitor(_encoder, callback=_progress_bar)
            return self.do_http_request(  # type: ignore
                url_path,
                instance_uri=instance_uri,
                method=HTTPMethod.POST,
                timeout=1200,
                data=_monitor,
                headers=_headers,
                **kw,
            )

    def do_http_request_simple_ret(
        self,
        path: str,
        instance_uri: URI,
        method: str = HTTPMethod.GET,
        **kw: t.Any,
    ) -> t.Tuple[bool, str]:
        r = self.do_http_request(path, instance_uri, method, **kw)
        status = r.status_code == HTTPStatus.OK

        try:
            message = r.json()["message"]
        except Exception as e:
            message = r.text or str(e)

        return status, message

    @staticmethod
    @http_retry
    def do_http_request(
        path: str,
        instance_uri: URI,
        method: str = HTTPMethod.GET,
        timeout: int = _DEFAULT_TIMEOUT_SECS,
        headers: t.Optional[t.Dict[str, t.Any]] = None,
        disable_default_content_type: bool = False,
        **kw: t.Any,
    ) -> requests.Response:
        _url = f"{instance_uri.instance}/api/{SW_API_VERSION}/{path.lstrip('/')}"
        _headers = {
            "Authorization": instance_uri.sw_token,
        }
        if not disable_default_content_type:
            _headers["Content-Type"] = "application/json"

        if headers is not None:
            _headers.update(headers)

        use_raise = kw.pop("use_raise", False)
        ignore_status_codes = kw.pop("ignore_status_codes", [])

        r = requests.request(
            method,
            _url,
            timeout=timeout,
            verify=False,
            headers=_headers,
            **kw,
        )
        wrap_sw_error_resp(
            r,
            path,
            use_raise=use_raise,
            silent=True,
            ignore_status_codes=ignore_status_codes,
        )
        return r

    def parse_pager(self, resp: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        _d = resp["data"]
        return dict(
            total=_d["total"],
            current=_d["size"],
            remain=_d["total"] - _d["size"],
        )

    def _fetch_bundle_info(self, uri: URI, typ: str) -> t.Dict[str, t.Any]:
        _manifest: t.Dict[str, t.Any] = {
            "uri": uri.full_uri,
            "project": uri.project,
            "name": uri.object.name,
        }

        if uri.object.version:
            # TODO: add manifest, lock(runtime), model/dataset/runtime yaml info by controller api
            _manifest["version"] = uri.object.version
            _info = self._fetch_bundle_version_info(uri, typ)
            _manifest[CREATED_AT_KEY] = self.fmt_timestamp(_info["createdTime"])
            _manifest["size"] = _info["files"][0]["size"]
        else:
            _manifest["history"] = self._fetch_bundle_history(
                name=uri.object.name,
                project_uri=uri,
                typ=typ,
            )[0]
        return _manifest

    def _fetch_bundle_version_info(self, uri: URI, typ: str) -> t.Dict[str, t.Any]:
        r = self.do_http_request(
            f"/project/{uri.project}/{typ}/{uri.object.name}",
            method=HTTPMethod.GET,
            instance_uri=uri,
            params={"versionName": uri.object.version},
        ).json()
        return r["data"]  # type: ignore

    def _fetch_bundle_history(
        self,
        name: str,
        project_uri: URI,
        typ: str,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        r = self.do_http_request(
            f"/project/{project_uri.project}/{typ}/{name}/version",
            instance_uri=project_uri,
            method=HTTPMethod.GET,
            params={"pageNum": page, "pageSize": size},
        ).json()

        _history = []
        for _h in r["data"]["list"]:
            # TODO: add manifest
            _history.append(
                {
                    "id": _h["id"],
                    "name": name,
                    "version": _h["name"],
                    "size": self.get_bundle_size_from_resp(typ, _h),
                    CREATED_AT_KEY: self.fmt_timestamp(_h["createdTime"]),
                    "is_removed": _h.get("is_removed", False),
                }
            )

        return _history, self.parse_pager(r)

    def _fetch_bundle_all_list(
        self,
        project_uri: URI,
        uri_typ: str,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filter_dict: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        filter_dict = filter_dict or {}
        _params = {"pageNum": page, "pageSize": size}
        _params.update(filter_dict)
        r = self.do_http_request(
            f"/project/{project_uri.project}/{uri_typ}",
            params=_params,
            instance_uri=project_uri,
        ).json()
        objects = {}

        _page = page
        _size = size
        if filter_dict.get("latest") is not None:
            _page = 1
            _size = 1

        for o in r["data"]["list"]:
            _name = f"[{o['id']}] {o['name']}"
            objects[_name] = self._fetch_bundle_history(
                name=o["id"],
                project_uri=project_uri,
                typ=uri_typ,
                page=_page,
                size=_size,
            )[0]

        return objects, self.parse_pager(r)

    def get_bundle_size_from_resp(self, typ: str, item: t.Dict) -> int:
        default_size = 0
        size = item.get("size", default_size)
        if size:
            return int(size)
        meta_str = item.get("meta", False)
        if not meta_str:
            return default_size

        meta = yaml.safe_load(meta_str)
        if not isinstance(meta, dict):
            return default_size

        if typ == "dataset":
            return int(
                meta.get("dataset_summary", {}).get("blobs_byte_size", default_size)
            )
        if typ == "runtime":
            # no size info in meta for now
            return default_size

        return default_size


class CloudBundleModelMixin(CloudRequestMixed):
    def info(self) -> t.Dict[str, t.Any]:
        uri: URI = self.uri  # type: ignore
        return self._fetch_bundle_info(uri, uri.object.typ)

    @ignore_error(({}, {}))
    def history(
        self, page: int = DEFAULT_PAGE_IDX, size: int = DEFAULT_PAGE_SIZE
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        uri: URI = self.uri  # type: ignore
        return self._fetch_bundle_history(
            name=uri.object.name,
            project_uri=uri,
            typ=uri.object.typ,
            page=page,
            size=size,
        )

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: remove specific version
        # TODO: add force flag
        uri: URI = self.uri  # type: ignore
        return self.do_http_request_simple_ret(
            f"/project/{uri.project}/{uri.object.typ}/{uri.object.name}",
            method=HTTPMethod.DELETE,
            instance_uri=uri,
        )

    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: recover specific version
        # TODO: add force flag
        uri: URI = self.uri  # type: ignore
        return self.do_http_request_simple_ret(
            f"/project/{uri.project}/{uri.object.typ}/{uri.object.name}/recover",
            method=HTTPMethod.PUT,
            instance_uri=uri,
        )

    def list_tags(self) -> t.List[str]:
        raise NoSupportError("no support list tags for dataset in the cloud instance")

    def add_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        raise NoSupportError("no support add tags for dataset in the cloud instance")

    def remove_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        raise NoSupportError("no support remove tags for dataset in the cloud instance")



import os
import typing as t
from http import HTTPStatus
from pathlib import Path
from concurrent.futures import wait, ThreadPoolExecutor

import yaml
from rich.progress import (
    TaskID,
    Progress,
    BarColumn,
    TextColumn,
    SpinnerColumn,
    TimeElapsedColumn,
    TotalFileSizeColumn,
    TransferSpeedColumn,
)

from starwhale.utils import console, now_str, load_yaml
from starwhale.consts import (
    FileDesc,
    FileNode,
    HTTPMethod,
    CREATED_AT_KEY,
    VERSION_PREFIX_CNT,
    STANDALONE_INSTANCE,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.base.tag import StandaloneTag
from starwhale.base.uri import URI
from starwhale.utils.fs import ensure_dir, ensure_file
from starwhale.base.type import URIType, InstanceType, get_bundle_type_by_uri
from starwhale.base.cloud import CloudRequestMixed
from starwhale.utils.error import NotFoundError, NoSupportError, FieldTypeOrValueError
from starwhale.utils.config import SWCliConfigMixed
from starwhale.base.blob.store import LocalFileStore
from starwhale.core.model.store import ModelStorage
from starwhale.core.dataset.store import DatasetStorage
from starwhale.core.runtime.store import RuntimeStorage
from starwhale.base.uricomponents.project import Project
from starwhale.base.uricomponents.resource import Resource, ResourceType

TMP_FILE_BUFSIZE = 8192

_query_param_map = {
    URIType.DATASET: "swds",
    URIType.MODEL: "swmp",
    URIType.RUNTIME: "runtime",
}


class _UploadPhase:
    MANIFEST = "MANIFEST"
    BLOB = "BLOB"
    END = "END"
    CANCEL = "CANCEL"


class BundleCopy(CloudRequestMixed):
    def __init__(
        self,
        src_uri: str,
        dest_uri: str,
        typ: str,
        force: bool = False,
        **kw: t.Any,
    ) -> None:
        self.src_resource = Resource(src_uri, typ=ResourceType[typ])
        self.src_uri = self.src_resource.to_uri()
        if self.src_uri.instance_type == InstanceType.CLOUD:
            p = kw.get("dest_local_project_uri")
            project = p and Project(p) or None
            dest_uri = self.src_uri.object.name if dest_uri.strip() == "." else dest_uri
        else:
            project = None

        try:
            self.dest_uri = Resource(
                dest_uri, typ=ResourceType[typ], project=project
            ).to_uri()
        except Exception as e:
            if str(e).startswith("invalid uri"):
                self.dest_uri = Project(dest_uri).to_uri()
            else:
                raise

        self.typ = typ
        self.force = force
        self.bundle_name = self.src_uri.object.name
        self.bundle_version = self._guess_bundle_version()
        self.field_flag = _query_param_map[self.typ]
        self.field_value = f"{self.bundle_name}:{self.bundle_version}"

        self.kw = kw

        self._sw_config = SWCliConfigMixed()
        self._do_validate()
        self._object_store = LocalFileStore()

    def _guess_bundle_version(self) -> str:
        if self.src_uri.instance_type == InstanceType.CLOUD:
            return self.src_uri.object.version
        else:
            if self.typ == URIType.DATASET:
                _v = DatasetStorage(self.src_uri).id
            elif self.typ == URIType.MODEL:
                _v = ModelStorage(self.src_uri).id
            elif self.typ == URIType.RUNTIME:
                _v = RuntimeStorage(self.src_uri).id
            else:
                raise NoSupportError(self.typ)

            return _v or self.src_uri.object.version

    def _do_validate(self) -> None:
        if self.typ not in (URIType.DATASET, URIType.MODEL, URIType.RUNTIME):
            raise NoSupportError(f"{self.typ} copy does not work")

        if self.bundle_version == "":
            raise Exception(f"must specify version src:({self.bundle_version})")

    def _check_cloud_obj_existed(self, instance_uri: URI) -> bool:
        # TODO: add more params for project
        # TODO: tune controller api, use unified params name
        ok, _ = self.do_http_request_simple_ret(
            path=self._get_remote_bundle_api_url(for_head=True),
            method=HTTPMethod.HEAD,
            instance_uri=instance_uri,
            ignore_status_codes=[HTTPStatus.NOT_FOUND],
        )
        return ok

    def _get_versioned_resource_path(self, uri: URI) -> Path:
        if uri.instance_type != InstanceType.STANDALONE:
            raise NoSupportError(f"{uri} to get target dir path")

        resource_name = uri.object.name or self.bundle_name
        return (
            self._sw_config.rootdir
            / uri.project
            / self.typ
            / resource_name
            / self.bundle_version[:VERSION_PREFIX_CNT]
            / f"{self.bundle_version}{get_bundle_type_by_uri(self.typ)}"
        )

    def _check_version_existed(self, uri: URI) -> bool:
        if uri.instance_type == InstanceType.CLOUD:
            return self._check_cloud_obj_existed(uri)
        else:
            return self._get_versioned_resource_path(uri).exists()

    def _get_remote_bundle_console_url(self, with_version: bool = True) -> str:
        if self.src_uri.instance_type == InstanceType.CLOUD:
            remote = self.src_uri
            resource_name = self.src_uri.object.name
        else:
            remote = self.dest_uri
            resource_name = self.dest_uri.object.name or self.src_uri.object.name

        url = f"{remote.instance}/projects/{remote.project}/{self.typ}s/{resource_name}"
        if with_version:
            url = f"{url}/versions/{self.src_uri.object.version}/overview"
        return url

    def _get_remote_bundle_api_url(self, for_head: bool = False) -> str:
        version = self.src_uri.object.version
        if not version:
            raise FieldTypeOrValueError(
                f"cannot fetch version from src uri:{self.src_uri}"
            )

        if self.src_uri.instance_type == InstanceType.CLOUD:
            project = self.src_uri.project
            resource_name = self.src_uri.object.name
        else:
            project = self.dest_uri.project
            resource_name = self.dest_uri.object.name or self.src_uri.object.name

        if not resource_name:
            raise FieldTypeOrValueError(
                f"cannot fetch {self.typ} resource name from src_uri({self.src_uri}) or dest_uri({self.dest_uri})"
            )

        base = [f"/project/{project}/{self.typ}/{resource_name}/version/{version}"]
        if not for_head:
            # uri for head request contains no 'file'
            base.append("file")

        return "/".join(base)

    def _do_upload_bundle_tar(self, progress: Progress) -> None:
        file_path = self._get_versioned_resource_path(self.src_uri)
        task_id = progress.add_task(
            f":bowling: upload {file_path.name}",
            total=file_path.stat().st_size,
        )

        self.do_multipart_upload_file(
            url_path=self._get_remote_bundle_api_url(),
            file_path=file_path,
            instance_uri=self.dest_uri,
            fields={
                self.field_flag: self.field_value,
                "project": self.dest_uri.project,
                "force": "1" if self.force else "0",
            },
            use_raise=True,
            progress=progress,
            task_id=task_id,
        )

    def _do_download_bundle_tar(self, progress: Progress) -> None:
        file_path = self._get_versioned_resource_path(self.dest_uri)
        ensure_dir(os.path.dirname(file_path))
        task_id = progress.add_task(f":bowling: download to {file_path}...")

        self.do_download_file(
            url_path=self._get_remote_bundle_api_url(),
            dest_path=file_path,
            instance_uri=self.src_uri,
            params={
                self.field_flag: self.field_value,
                "project": self.src_uri.project,
            },
            progress=progress,
            task_id=task_id,
        )

    def do(self) -> None:
        remote_url = self._get_remote_bundle_console_url()
        if self._check_version_existed(self.dest_uri) and not self.force:
            console.print(f":tea: {remote_url} was already existed, skip copy")
            return

        if not self._check_version_existed(self.src_uri):
            raise NotFoundError(str(self.src_uri))

        console.print(f":construction: start to copy {self.src_uri} -> {self.dest_uri}")

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            TimeElapsedColumn(),
            TotalFileSizeColumn(),
            TransferSpeedColumn(),
            console=console,
            refresh_per_second=0.2,
        ) as progress:
            if self.src_uri.instance_type == InstanceType.STANDALONE:
                if self.typ == URIType.MODEL:
                    self._do_upload_bundle_dir(progress)
                elif self.typ == URIType.RUNTIME:
                    self._do_upload_bundle_tar(progress)
                else:
                    raise NoSupportError(
                        f"no support to copy {self.typ} from standalone to server"
                    )
            else:
                if self.typ == URIType.MODEL:
                    self._do_download_bundle_dir(progress)
                elif self.typ == URIType.RUNTIME:
                    self._do_download_bundle_tar(progress)
                else:
                    raise NoSupportError(
                        f"no support to copy {self.typ} from server to standalone"
                    )

                _dest_uri = URI.capsulate_uri(
                    instance=STANDALONE_INSTANCE,
                    project=self.dest_uri.project,
                    obj_type=self.typ,
                    obj_name=self.dest_uri.object.name or self.bundle_name,
                    obj_ver=self.bundle_version,
                )
                StandaloneTag(_dest_uri).add_fast_tag()
                self._update_manifest(
                    self._get_versioned_resource_path(_dest_uri),
                    {CREATED_AT_KEY: now_str()},
                )
        console.print(f":tea: console url of the remote bundle: {remote_url}")

    def upload_files(self, workdir: Path) -> t.Iterator[FileNode]:
        raise NotImplementedError

    def download_files(self, workdir: Path) -> t.Iterator[FileNode]:
        raise NotImplementedError

    def _do_download_bundle_dir(self, progress: Progress) -> None:
        _workdir = self._get_versioned_resource_path(self.dest_uri)

        def _download(_tid: TaskID, fd: FileNode) -> None:
            if fd.signature:
                f = self._object_store.get(fd.signature)
                if f is not None and f.exists():
                    f.link(_workdir / fd.path / fd.name)
                    progress.update(_tid, completed=fd.size)
                    return

            self.do_download_file(
                # TODO: use /project/{self.typ}/pull api
                url_path=self._get_remote_bundle_api_url(),
                dest_path=fd.path,
                instance_uri=self.src_uri,
                params={
                    "desc": fd.file_desc.name,
                    "partName": fd.name,
                    "signature": fd.signature,
                },
                progress=progress,
                task_id=_tid,
            )
            # put the downloaded file to object store for cache usage
            self._object_store.link(fd.path, fd.signature)

        _manifest_path = _workdir / DEFAULT_MANIFEST_NAME
        _tid = progress.add_task(f":arrow_down: {DEFAULT_MANIFEST_NAME}")
        _download(
            _tid,
            FileNode(
                path=_manifest_path,
                name=DEFAULT_MANIFEST_NAME,
                signature="",
                file_desc=FileDesc.MANIFEST,
                size=0,
            ),
        )

        for _f in self.download_files(workdir=_workdir):
            _tid = progress.add_task(
                f":arrow_down: {_f.name}",
                total=float(_f.size),
            )
            if not _f.path.exists() or self.force:
                _download(_tid, _f)

    def _do_ubd_bundle_prepare(
        self,
        progress: t.Optional[Progress],
        workdir: Path,
        url_path: str,
    ) -> t.Any:
        manifest_path = workdir / DEFAULT_MANIFEST_NAME
        if progress is None:
            task_id = TaskID(0)
        else:
            task_id = progress.add_task(
                f":arrow_up: {manifest_path.name}",
                total=manifest_path.stat().st_size,
            )

        # TODO: use rich progress
        r = self.do_multipart_upload_file(
            url_path=url_path,
            file_path=manifest_path,
            instance_uri=self.dest_uri,
            params={
                self.field_flag: self.field_value,
                "phase": _UploadPhase.MANIFEST,
                "desc": FileDesc.MANIFEST.name,
                "project": self.dest_uri.project,
                "force": "1" if self.force else "0",
            },
            use_raise=True,
            progress=progress,
            task_id=task_id,
        )
        return r.json().get("data", {})

    def _do_ubd_blobs(
        self,
        progress: t.Optional[Progress],
        workdir: Path,
        upload_id: str,
        url_path: str,
        existed_files: t.Optional[t.List] = None,
    ) -> None:
        existed_files = existed_files or []

        # TODO: add retry deco
        def _upload_blob(_tid: TaskID, fd: FileNode) -> None:
            if not fd.path.exists():
                raise NotFoundError(f"{fd.path} not found")

            if progress is not None:
                progress.update(_tid, visible=True)
            self.do_multipart_upload_file(
                url_path=url_path,
                file_path=fd.path,
                instance_uri=self.dest_uri,
                params={
                    self.field_flag: self.field_value,
                    "phase": _UploadPhase.BLOB,
                    "uploadId": upload_id,
                    "partName": fd.name,
                    "signature": fd.signature,
                    "desc": fd.file_desc.name,
                },
                use_raise=True,
                progress=progress,
                task_id=_tid,
            )

        _p_map = {}
        for _id, _f in enumerate(self.upload_files(workdir=workdir)):
            if existed_files and _f.signature in existed_files:
                continue

            if progress is None:
                _tid = TaskID(_id)
            else:
                _tid = progress.add_task(
                    f":arrow_up: {_f.path.name}",
                    total=float(_f.size),
                    visible=False,
                )
            _p_map[_tid] = _f
        with ThreadPoolExecutor(
            max_workers=int(os.environ.get("SW_BUNDLE_COPY_THREAD_NUM", "5"))
        ) as executor:
            futures = [
                executor.submit(_upload_blob, _tid, _file_desc)
                for _tid, _file_desc in _p_map.items()
            ]
            # TODO throw errors
            wait(futures)

    def _do_ubd_end(self, upload_id: str, url_path: str, ok: bool) -> None:
        phase = _UploadPhase.END if ok else _UploadPhase.CANCEL
        self.do_http_request(
            path=url_path,
            method=HTTPMethod.POST,
            instance_uri=self.dest_uri,
            data={
                self.field_flag: self.field_value,
                "project": self.dest_uri.project,
                "phase": phase,
                "uploadId": upload_id,
            },
            use_raise=True,
            disable_default_content_type=True,
        )

    def _do_upload_bundle_dir(
        self,
        progress: t.Optional[Progress] = None,
        workdir: t.Optional[Path] = None,
    ) -> None:
        workdir = workdir or self._get_versioned_resource_path(self.src_uri)
        url_path = self._get_remote_bundle_api_url()

        res_data = self._do_ubd_bundle_prepare(
            progress=progress,
            workdir=workdir,
            url_path=url_path,
        )
        upload_id: str = res_data.get("uploadId", "")
        if not upload_id:
            raise Exception("upload id is empty")
        exists_files: list = res_data.get("existed", [])
        try:
            self._do_ubd_blobs(
                progress=progress,
                workdir=workdir,
                upload_id=upload_id,
                url_path=url_path,
                existed_files=exists_files,
            )
        except Exception as e:
            console.print(
                f":confused_face: when upload blobs, we meet Exception{e}, will cancel upload"
            )
            self._do_ubd_end(upload_id=upload_id, url_path=url_path, ok=False)
            raise
        else:
            self._do_ubd_end(upload_id=upload_id, url_path=url_path, ok=True)

    @staticmethod
    def _update_manifest(workdir: Path, patch: t.Dict[str, t.Any]) -> None:
        file = workdir / DEFAULT_MANIFEST_NAME
        # downloaded runtime is a tarball
        if not file.exists():
            return
        manifest = load_yaml(file)
        manifest.update(patch)
        ensure_file(file, yaml.safe_dump(manifest, default_flow_style=False))



from __future__ import annotations

import json
import typing as t
from abc import ABCMeta, abstractmethod
from pathlib import Path

import yaml

from starwhale.utils import validate_obj_name
from starwhale.consts import (
    HTTPMethod,
    CREATED_AT_KEY,
    DEFAULT_PROJECT,
    RECOVER_DIRNAME,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    SHORT_VERSION_CNT,
)
from starwhale.base.uri import URI
from starwhale.utils.fs import move_dir, ensure_dir, get_path_created_time
from starwhale.base.type import URIType, InstanceType
from starwhale.base.cloud import CloudRequestMixed
from starwhale.utils.http import ignore_error
from starwhale.utils.error import NoSupportError
from starwhale.utils.config import SWCliConfigMixed

_SHOW_ALL = 100


class ProjectObjType:
    MODEL = "model"
    DATASET = "dataset"
    RUNTIME = "runtime"


class Project(metaclass=ABCMeta):
    def __init__(self, uri: URI) -> None:
        self.uri = uri
        self.name = uri.project.lower()
        self.sw_config = SWCliConfigMixed()

    @abstractmethod
    def create(self) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def info(self) -> t.Dict[str, t.Any]:
        raise NotImplementedError

    @abstractmethod
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def recover(self) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @classmethod
    def list(
        cls,
        instance_uri: str = "",
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _uri = URI(instance_uri, expected_type=URIType.INSTANCE)
        if _uri.instance_type == InstanceType.STANDALONE:
            return StandaloneProject.list()
        elif _uri.instance_type == InstanceType.CLOUD:
            return CloudProject.list(instance_uri, page, size)  # type: ignore
        else:
            raise NoSupportError(f"{instance_uri}")

    @classmethod
    def get_project(cls, project_uri: URI) -> Project:
        if project_uri.instance_type == InstanceType.STANDALONE:
            return StandaloneProject(project_uri)
        elif project_uri.instance_type == InstanceType.CLOUD:
            return CloudProject(project_uri)
        else:
            raise NoSupportError(f"{project_uri}")


class StandaloneProject(Project):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.loc = Path(self.sw_config.rootdir / self.name)
        self.recover_dir = Path(self.sw_config.rootdir / RECOVER_DIRNAME)

    def create(self) -> t.Tuple[bool, str]:
        ok, reason = validate_obj_name(self.name)
        if not ok:
            return ok, reason

        if self.loc.exists():
            return False, f"{self.loc} was already existed"

        try:
            ensure_dir(self.loc)
        except Exception as e:
            return False, str(e)
        else:
            return True, f"{self.loc} created"

    def info(self) -> t.Dict[str, t.Any]:
        if not self.loc.exists():
            return {}

        # TODO: add more project info
        return {
            "name": self.name,
            "location": str(self.loc),
            CREATED_AT_KEY: get_path_created_time(self.loc),
        }

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        if self.name == DEFAULT_PROJECT:
            return False, f"prohibit to remove {self.name}"

        ensure_dir(self.recover_dir)
        return move_dir(self.loc, self.recover_dir / self.name, force)

    def recover(self) -> t.Tuple[bool, str]:
        if self.name == DEFAULT_PROJECT:
            return False, f"prohibit to recover {self.name}"

        return move_dir(self.recover_dir / self.name, self.loc)

    @classmethod
    def list(
        cls,
        instance_uri: str = "",
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        sw = SWCliConfigMixed()

        if not sw.rootdir.exists():
            return [], {}

        rt = []
        for pdir in sw.rootdir.iterdir():
            # TODO: add more project details
            valid, _ = validate_obj_name(pdir.name)
            if not valid:
                continue

            rt.append(
                {
                    "name": pdir.name,
                    "location": str(pdir),
                    CREATED_AT_KEY: get_path_created_time(pdir),
                }
            )

        return rt, {}


class CloudProject(Project, CloudRequestMixed):
    def create(self) -> t.Tuple[bool, str]:
        return self.do_http_request_simple_ret(
            "/project",
            method=HTTPMethod.POST,
            instance_uri=self.uri,
            data=json.dumps({"projectName": self.name}),
        )

    def recover(self) -> t.Tuple[bool, str]:
        return self.do_http_request_simple_ret(
            f"/project/{self.name}/action/recover",
            method=HTTPMethod.POST,
            instance_uri=self.uri,
        )

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.do_http_request_simple_ret(
            f"/project/{self.name}",
            method=HTTPMethod.DELETE,
            instance_uri=self.uri,
            data=json.dumps({"force": int(force)}),
        )

    @classmethod
    @ignore_error(([], {}))
    def list(
        cls,
        instance_uri: str = "",
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        crm = CloudRequestMixed()
        uri = URI(instance_uri, expected_type=URIType.INSTANCE)
        r = crm.do_http_request(
            "/project",
            params={"pageNum": page, "pageSize": size, "ownerName": uri.user_name},
            instance_uri=uri,
        ).json()

        projects = []
        for _p in r["data"]["list"]:
            owner = _p["owner"]["name"]
            projects.append(
                dict(
                    id=_p["id"],
                    name=_p["name"],
                    created_at=crm.fmt_timestamp(_p["createdTime"]),  # type: ignore
                    is_default=_p.get("isDefault", 0),
                    owner=owner,
                )
            )
        return projects, crm.parse_pager(r)

    @ignore_error({})
    def info(self) -> t.Dict[str, t.Any]:
        r = self.do_http_request(
            f"/project/{self.name}",
            method=HTTPMethod.GET,
            instance_uri=self.uri,
        )
        # TODO: add more project details
        return {
            "name": self.name,
            CREATED_AT_KEY: self.fmt_timestamp(r.json()["data"]["createdTime"]),
            "location": r.url,
            "models": self._fetch_project_objects(ProjectObjType.MODEL),
            "datasets": self._fetch_project_objects(ProjectObjType.DATASET),
        }

    @ignore_error([])
    def _fetch_model_files(self, mid: int) -> t.List:
        r = self.do_http_request(
            f"/project/{self.name}/model/{mid}", instance_uri=self.uri
        )
        return r.json()["data"]["files"]  # type: ignore

    @ignore_error([])
    def _fetch_project_objects(
        self, typ: str, versions_size: int = 10
    ) -> t.List[t.Dict[str, t.Any]]:
        r = self.do_http_request(
            f"/project/{self.name}/{typ}",
            params={"pageSize": _SHOW_ALL},
            instance_uri=self.uri,
        )

        ret = []
        for _m in r.json()["data"]["list"]:
            _m[CREATED_AT_KEY] = self.fmt_timestamp(_m.pop("createdTime"))
            _m.pop("owner", None)

            mvr = self.do_http_request(
                f"/project/{self.name}/{typ}/{_m['id']}/version",
                params={"pageSize": versions_size},
                instance_uri=self.uri,
            )
            versions = []
            for _v in mvr.json()["data"]["list"]:
                _v["short_name"] = _v["name"][:SHORT_VERSION_CNT]
                _v[CREATED_AT_KEY] = self.fmt_timestamp(_v.pop("createdTime"))
                _v.pop("owner", None)
                if typ == ProjectObjType.DATASET:
                    _v["meta"] = yaml.safe_load(_v["meta"])
                versions.append(_v)

            _m["latest_versions"] = versions
            if typ == ProjectObjType.MODEL:
                _m["files"] = self._fetch_model_files(_m["id"])
            ret.append(_m)

        return ret



from __future__ import annotations

import os
import typing as t
import inspect
import tempfile
from pathlib import Path

import yaml

from starwhale.utils import disable_progress_bar
from starwhale.consts import DefaultYAMLName
from starwhale.utils.fs import ensure_file
from starwhale.consts.env import SWEnv

_path_T = t.Union[str, Path]


def build(
    evaluation_handler: t.Any,
    workdir: t.Optional[_path_T] = None,
    name: t.Optional[str] = None,
    project_uri: str = "",
    desc: str = "",
    remote_project_uri: t.Optional[str] = None,
) -> None:
    """Build Starwhale Model Package.

    In common case, you may call `build` function in your experiment scripts.`build` function is a shortcut for the `swcli model build` command.

    Arguments:
        evaluation_handler: (str, object, required) The evaluation handler supports object(function, class or module) or str(example: "to.path.module:name").
        name: (str, optional) The name of Starwhale Model Package, default is the current work dir.
        workdir: (str, Pathlib.Path, optional) The path of the rootdir. The default workdir is the current working dir.
        desc: (str, optional) The description of the Starwhale Model Package.
        project_uri: (str, optional) The project uri of the Starwhale Model Package. If the argument is not specified,
            the project_uri is the config value of `swcli project select` command.
        remote_project_uri: (str, optional) The destination project uri(cloud://remote-instance/project/starwhale) of the Starwhale Model Package

    Examples:
    ```python
    from starwhale import model

    # class handler
    from .user.code.evaluator import ExamplePipelineHandler
    model.build(ExamplePipelineHandler)

    # function handler
    from .user.code.evaluator import predict_image
    model.build(predict_image)

    # module handler, @step decorates function in this module
    from .user.code import evaluator
    model.build(evaluator)

    # str handler
    model.build("user.code.evaluator:ExamplePipelineHandler")
    model.build("user.code.evaluator:predict_image")
    model.build("user.code.evaluator")
    ```

    Returns:
        None.
    """
    from starwhale.core.model.view import ModelTermView

    if workdir is None:
        workdir = Path.cwd()
    else:
        workdir = Path(workdir)

    name = name or workdir.name

    yaml_path = _render_model_yaml(
        handler=evaluation_handler,
        workdir=workdir,
        name=name,
        desc=desc,
    )
    try:
        with disable_progress_bar():
            ModelTermView.build(
                workdir=workdir,
                project=project_uri,
                yaml_path=yaml_path,
            )

            from starwhale import URI, URIType

            if remote_project_uri:
                remote_project_uri = URI(
                    remote_project_uri, expected_type=URIType.PROJECT
                ).full_uri
            elif os.getenv(SWEnv.instance_uri) and os.getenv(SWEnv.project):
                remote_project_uri = f"{os.getenv(SWEnv.instance_uri)}/project/{os.getenv(SWEnv.project)}"

            if remote_project_uri:
                ModelTermView.copy(
                    src_uri=f"{URI(project_uri).full_uri}/model/{name}/version/latest",
                    dest_uri=remote_project_uri,
                )
    finally:
        os.unlink(yaml_path)


def _ingest_obj_entrypoint_name(
    obj: t.Any,
    workdir: Path,
) -> str:
    source_path: t.Optional[_path_T] = inspect.getsourcefile(obj)
    if source_path is None:
        raise RuntimeError(f"failed to get source path for object: {obj}")
    source_path = Path(source_path).resolve().absolute()
    workdir = workdir.resolve().absolute()

    relative_path = str(source_path.relative_to(workdir))
    _parts = relative_path.split("/")
    module_import_path = ".".join([p.rsplit(".", 1)[0] for p in _parts if p])

    if inspect.ismodule(obj):
        return module_import_path
    else:
        return f"{module_import_path}:{obj.__qualname__}"


# TODO: remove _render_model_yaml function when the model.yaml is optional for model build
def _render_model_yaml(
    handler: t.Any,
    workdir: Path,
    name: str,
    desc: str,
) -> str:
    if not isinstance(handler, str):
        handler = _ingest_obj_entrypoint_name(handler, workdir)

    config = {
        "name": name,
        "run": {
            "handler": handler,
        },
        "desc": desc,
    }

    _, yaml_fpath = tempfile.mkstemp(suffix=DefaultYAMLName.MODEL)
    ensure_file(yaml_fpath, yaml.safe_dump(config, default_flow_style=False))
    return yaml_fpath



from __future__ import annotations

import io
import sys
import time
import typing as t
import inspect
import logging
import threading
from abc import ABCMeta, abstractmethod
from types import TracebackType
from pathlib import Path
from functools import wraps

import jsonlines

from starwhale.utils import now_str
from starwhale.consts import (
    RunStatus,
    CURRENT_FNAME,
    DecoratorInjectAttr,
    DEFAULT_EVALUATION_JOB_NAME,
)
from starwhale.base.uri import URI
from starwhale.utils.fs import ensure_dir, ensure_file
from starwhale.api._impl import wrapper
from starwhale.base.type import URIType, RunSubDirType
from starwhale.utils.log import StreamWrapper
from starwhale.api.service import Input, Output, Service
from starwhale.utils.error import NoSupportError, ParameterError, FieldTypeOrValueError
from starwhale.api._impl.job import step, AFTER_LOAD_HOOKS
from starwhale.core.eval.store import EvaluationStorage
from starwhale.core.job.context import Context
from starwhale.api._impl.dataset import Dataset
from starwhale.core.dataset.tabular import TabularDatasetRow

if t.TYPE_CHECKING:
    import loguru


class _LogType:
    SW = "starwhale"
    USER = "user"


_jl_writer: t.Callable[[Path], jsonlines.Writer] = lambda p: jsonlines.open(
    str((p).resolve()), mode="w"
)


def default_handler_hook(run_handler: str, func_or_cls: t.Any) -> None:
    module_name, _, func_or_cls_name = run_handler.partition(":")
    from loguru import logger

    # TODO: raise exception for @step, @predict or @evaluate in the Pipeline methods
    # TODO: step decorate auto inject some flags into function builtin fields
    # TODO: need to handle DecoratorInjectAttr.Evaluate ?
    if inspect.isfunction(func_or_cls):
        if getattr(func_or_cls, DecoratorInjectAttr.Predict, False) or getattr(
            func_or_cls, DecoratorInjectAttr.Step, False
        ):
            logger.debug(
                f"preload function-{func_or_cls_name} from module-{module_name}"
            )
        else:
            # TODO remove(or log warning) this exception when support multi handlers
            raise RuntimeError(
                f"preload function-{func_or_cls_name} does not use step or predict decorator"
            )
    elif inspect.isclass(func_or_cls):
        if issubclass(func_or_cls, PipelineHandler):
            ppl_func = getattr(func_or_cls, "ppl")
            cmp_func = getattr(func_or_cls, "cmp")
            step(job_name=DEFAULT_EVALUATION_JOB_NAME, task_num=2, name="ppl")(ppl_func)
            step(
                job_name=DEFAULT_EVALUATION_JOB_NAME,
                task_num=1,
                needs=["ppl"],
                name="cmp",
            )(cmp_func)
            logger.debug(f"preload class-{func_or_cls_name} from Pipeline")
        else:
            logger.debug(f"preload user custom class-{func_or_cls_name}")
    else:
        raise NoSupportError(f"failed to preload for {run_handler}")


AFTER_LOAD_HOOKS.setdefault("default_hook", default_handler_hook)


class PipelineHandler(metaclass=ABCMeta):
    def __init__(
        self,
        ppl_batch_size: int = 1,
        ignore_dataset_data: bool = False,
        ignore_error: bool = False,
        flush_result: bool = False,
        ppl_auto_log: bool = True,
        dataset_uris: t.Optional[t.List[str]] = None,
    ) -> None:
        self.ppl_batch_size = ppl_batch_size
        self.svc = Service()
        self.context = Context.get_runtime_context()

        self.dataset_uris = self.context.dataset_uris or dataset_uris or []

        # TODO: add args for compare result and label directly
        self.ignore_dataset_data = ignore_dataset_data
        self.ignore_error = ignore_error
        self.flush_result = flush_result
        self.ppl_auto_log = ppl_auto_log

        _logdir = EvaluationStorage.local_run_dir(
            self.context.project, self.context.version
        )
        _run_dir = (
            _logdir / RunSubDirType.RUNLOG / self.context.step / str(self.context.index)
        )
        self.status_dir = _run_dir / RunSubDirType.STATUS
        self.log_dir = _run_dir / RunSubDirType.LOG
        ensure_dir(self.status_dir)
        ensure_dir(self.log_dir)

        self.logger, self._sw_logger = self._init_logger(self.log_dir)
        self._stdout_changed = False
        self._stderr_changed = False
        self._orig_stdout = sys.stdout
        self._orig_stderr = sys.stderr
        # TODO: split status/result files
        self._timeline_writer = _jl_writer(self.status_dir / "timeline")

        # TODO: use EvaluationLogStore to refactor this?
        self.evaluation_store = wrapper.Evaluation(
            eval_id=self.context.version, project=self.context.project
        )
        self._monkey_patch()
        self._update_status(RunStatus.START)

    def _init_logger(
        self, log_dir: Path, rotation: str = "500MB"
    ) -> t.Tuple[loguru.Logger, loguru.Logger]:
        # TODO: remove logger first?
        # TODO: add custom log format, include daemonset pod name
        from loguru import logger as _logger

        # TODO: configure log rotation size
        _logger.add(
            log_dir / "{time}.log",
            rotation=rotation,
            backtrace=True,
            diagnose=True,
            serialize=True,
        )
        _logger.bind(
            type=_LogType.USER,
            task_id=self.context.index,
            job_id=self.context.version,
        )
        _sw_logger = _logger.bind(type=_LogType.SW)
        return _logger, _sw_logger

    def _monkey_patch(self) -> None:
        if not isinstance(sys.stdout, StreamWrapper) and isinstance(
            sys.stdout, io.TextIOWrapper
        ):
            sys.stdout = StreamWrapper(sys.stdout, self.logger, logging.INFO)  # type: ignore
            self._stdout_changed = True

        if not isinstance(sys.stderr, StreamWrapper) and isinstance(
            sys.stderr, io.TextIOWrapper
        ):
            sys.stderr = StreamWrapper(sys.stderr, self.logger, logging.WARN)  # type: ignore
            self._stderr_changed = True

    def __str__(self) -> str:
        return f"PipelineHandler status@{self.status_dir}, " f"log@{self.log_dir}"

    def __enter__(self) -> PipelineHandler:
        return self

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        trace: TracebackType,
    ) -> None:
        self._sw_logger.debug(
            f"execute {self.context.step}-{self.context.index} exit func..."
        )
        if value:  # pragma: no cover
            print(f"type:{type}, exception:{value}, traceback:{trace}")

        if self._stdout_changed:
            sys.stdout = self._orig_stdout
        if self._stderr_changed:
            sys.stderr = self._orig_stderr
        self._timeline_writer.close()

    @abstractmethod
    def ppl(self, data: t.Any, **kw: t.Any) -> t.Any:
        # TODO: how to handle each element is not equal.
        raise NotImplementedError

    @abstractmethod
    def cmp(self, *args: t.Any, **kw: t.Any) -> t.Any:
        raise NotImplementedError

    def _record_status(func):  # type: ignore
        @wraps(func)  # type: ignore
        def _wrapper(*args: t.Any, **kwargs: t.Any) -> None:
            self: PipelineHandler = args[0]
            self._sw_logger.info(
                f"start to run {func.__name__} function@{self.context.step}-{self.context.index} ..."  # type: ignore
            )
            self._update_status(RunStatus.RUNNING)
            try:
                func(*args, **kwargs)  # type: ignore
            except Exception as e:
                self._update_status(RunStatus.FAILED)
                self._sw_logger.exception(f"{func} abort, exception: {e}")
                raise
            else:
                self._update_status(RunStatus.SUCCESS)

        return _wrapper

    @_record_status  # type: ignore
    def _starwhale_internal_run_cmp(self) -> None:
        now = now_str()
        try:
            if self.ppl_auto_log:
                self.cmp(self.evaluation_store.get_results(deserialize=True))
            else:
                self.cmp()
        except Exception as e:
            self._sw_logger.exception(f"cmp exception: {e}")
            self._timeline_writer.write(
                {"time": now, "status": False, "exception": str(e)}
            )
            raise
        else:
            self._timeline_writer.write({"time": now, "status": True, "exception": ""})

    def _is_ppl_batch(self) -> bool:
        return self.ppl_batch_size > 1

    @_record_status  # type: ignore
    def _starwhale_internal_run_ppl(self) -> None:
        if not self.dataset_uris:
            raise FieldTypeOrValueError("context.dataset_uris is empty")
        join_str = "_#@#_"
        cnt = 0
        # TODO: user custom config batch size, max_retries
        for uri_str in self.dataset_uris:
            _uri = URI(uri_str, expected_type=URIType.DATASET)
            ds = Dataset.dataset(_uri, readonly=True)
            ds.make_distributed_consumption(session_id=self.context.version)
            dataset_info = ds.info
            cnt = 0
            for rows in ds.batch_iter(self.ppl_batch_size):
                _start = time.time()
                _exception = None
                _results: t.Any = b""
                try:
                    if self._is_ppl_batch():
                        _results = self.ppl(
                            [row.features for row in rows],
                            index=[row.index for row in rows],
                            index_with_dataset=[
                                f"{_uri.object}{join_str}{row.index}" for row in rows
                            ],
                            dataset_info=dataset_info,
                        )
                    else:
                        _results = [
                            self.ppl(
                                rows[0].features,
                                index=rows[0].index,
                                index_with_dataset=f"{_uri.object}{join_str}{rows[0].index}",
                                dataset_info=dataset_info,
                            )
                        ]
                except Exception as e:
                    _exception = e
                    self._sw_logger.exception(
                        f"[{[r.index for r in rows]}] data handle -> failed"
                    )
                    if not self.ignore_error:
                        self._update_status(RunStatus.FAILED)
                        raise
                else:
                    _exception = None

                for (_idx, _features), _result in zip(rows, _results):
                    cnt += 1
                    _idx_with_ds = f"{_uri.object}{join_str}{_idx}"

                    self._sw_logger.debug(
                        f"[{_idx_with_ds}] use {time.time() - _start:.3f}s, session-id:{self.context.version} @{self.context.step}-{self.context.index}"
                    )

                    self._timeline_writer.write(
                        {
                            "time": now_str(),
                            "status": _exception is None,
                            "exception": str(_exception),
                            "index": _idx,
                            "index_with_dataset": _idx_with_ds,
                        }
                    )

                    if self.ppl_auto_log:
                        if not self.ignore_dataset_data:
                            for artifact in TabularDatasetRow.artifacts_of(_features):
                                if artifact.link:
                                    artifact.clear_cache()
                        self.evaluation_store.log_result(
                            data_id=_idx_with_ds,
                            index=_idx,
                            result=_result,
                            ds_data={}
                            if self.ignore_dataset_data
                            else _features.copy(),  # drop DataRow._Features type, keep dict type for features
                            serialize=True,
                        )

        if self.flush_result and self.ppl_auto_log:
            self.evaluation_store.flush_result()

        self._sw_logger.info(
            f"{self.context.step}-{self.context.index} handled {cnt} data items for dataset {self.dataset_uris}"
        )

    def _update_status(self, status: str) -> None:
        fpath = self.status_dir / CURRENT_FNAME
        ensure_file(fpath, status)

    def add_api(
        self, input: Input, output: Output, func: t.Callable, name: str
    ) -> None:
        self.svc.add_api(input, output, func, name)

    def serve(self, addr: str, port: int) -> None:
        self.svc.serve(addr, port)


# TODO: add flush, get_summary functions?
class EvaluationLogStore:
    _instance_holder = threading.local()
    _lock = threading.Lock()

    def __init__(self, id: str, project: str) -> None:
        self.id = id
        self.project = project

        self._datastore = wrapper.Evaluation(eval_id=self.id, project=self.project)

    def __str__(self) -> str:
        return f"Evaluation: id({self.id}), project({self.project})"

    __repr__ = __str__

    @classmethod
    def _get_instance(cls) -> EvaluationLogStore:
        with cls._lock:
            _inst: t.Optional[EvaluationLogStore]
            try:
                _inst = cls._instance_holder.value  # type: ignore
            except AttributeError:
                _inst = None

            if _inst is not None:
                return _inst

            context = Context.get_runtime_context()

            _inst = cls(
                id=context.version,
                project=context.project,
            )

            cls._instance_holder.value = _inst
            return _inst

    @classmethod
    def log(
        cls, category: str, id: t.Union[str, int], metrics: t.Dict[str, t.Any]
    ) -> None:
        """Log metrics into the Starwhale Datastore.

        Arguments:
            category: [str, required] The category of the log records.
            id: [Union[str, int], required] The unique id of the record.
            metrics: [Dict, required] The metrics dict.

        Examples:
        ```python
        from starwhale import evaluation

        evaluation.log("label/1", 1, {"loss": 0.99, "accuracy": 0.98})
        evaluation.log("ppl", "1", {"a": "test", "b": 1})

        Returns:
            None
        ```
        """
        # TODO: support pickle serialize
        el = cls._get_instance()
        el._log(category, id, metrics)

    def _log(
        self, category: str, id: t.Union[str, int], metrics: t.Dict[str, t.Any]
    ) -> None:
        self._datastore.log(table_name=category, id=id, **metrics)

    @classmethod
    def log_summary(cls, *args: t.Any, **kw: t.Any) -> None:
        """Log info into the Starwhale Datastore summary table.

        Arguments:
            *args: [dict, optional] use dicts to store summary info.
            **kwargs: [optional] use kv to store summary info.

        Examples:
        ```python
        from starwhale import evaluation

        evaluation.log_summary(loss=0.99)
        evaluation.log_summary(loss=0.99, accuracy=0.99)
        evaluation.log_summary({"loss": 0.99, "accuracy": 0.99})
        ```

        Returns:
            None
        """
        metrics = {}
        if len(args) == 0:
            metrics = kw
        elif len(args) == 1:
            if len(kw) > 0:
                raise ParameterError(
                    f"Args({args}) and kwargs({kw}) are specified at the same time"
                )
            if not isinstance(args[0], dict):
                raise ParameterError(f"Args({args[0]}) is not dict type")

            metrics = args[0]
        else:
            raise ParameterError(f"The number of args({args}) is greater than one")

        metrics.update(kw)
        el = cls._get_instance()
        el._log_summary(metrics)

    def _log_summary(self, metrics: t.Dict) -> None:
        self._datastore.log_metrics(metrics)

    @classmethod
    def iter(
        cls,
        category: str,
    ) -> t.Iterator:
        """Get an iterator for the log data of the specified category.

        Arguments:
            category: [str, required] The category of the log records.

        Examples:
        ```python
        from starwhale import evaluation

        results = [data for data in evaluation.iter("label/0")]
        ```

        Returns:
            An iterator.
        """
        # TODO: support batch_size
        el = cls._get_instance()
        return el._iter(category)

    def _iter(self, category: str) -> t.Iterator:
        return self._datastore.get(category)


def predict(*args: t.Any, **kw: t.Any) -> t.Any:
    """Defines a predict function.

    This function can be used as a decorator to define an evaluation-predict function that maps the dataset data into
    the different predict workers.

    Arguments:
        datasets: [Union[List[str], None], optional] Use the datasets rows as the input data for the predict function.
        resources: [Dict, optional] Resources for the predict task, such as memory, gpu etc. Current only supports
            the cloud instance.
        concurrency: [int, optional] The concurrency of the predict tasks. Default is 1.
        task_num: [int, optional] The number of the predict tasks. Default is 2.
        batch_size: [int, optional] Number of samples per batch. Default is 1.
        fail_on_error: [bool, optional] Fast fail on the exceptions in the predict function. Default is True.
        auto_log: [bool, optional] Auto log the return values of the predict function and the according dataset rows. Default is True.

    Examples:
    ```python

    from starwhale import evaluation

    @evaluation.predict
    def predict_image(data):
        ...

    @evaluation.predict(
        dataset="mnist/version/latest",
        batch_size=32,
        task_num=4,
        auto_log=True,
    )
    def predict_batch_images(batch_data)
        ...
    ```

    Returns:
        The decorated function.
    """

    # TODO: support runtime

    if len(args) == 1 and len(kw) == 0 and callable(args[0]):
        return predict()(args[0])
    else:

        def _wrap(func: t.Callable) -> t.Any:
            _register_predict(func, **kw)
            setattr(func, DecoratorInjectAttr.Predict, True)
            return func

        return _wrap


_registered_predict_func = threading.local()


def _register_predict(
    func: t.Callable,
    datasets: t.Optional[t.List[str]] = None,
    resources: t.Optional[t.Dict[str, t.Any]] = None,
    concurrency: int = 1,
    task_num: int = 2,
    batch_size: int = 1,
    fail_on_error: bool = True,
    auto_log: bool = True,
) -> None:
    from .job import step

    try:
        val = _registered_predict_func.value
    except AttributeError:
        val = None

    if val is not None:
        raise RuntimeError("predict function can only be called once")

    _registered_predict_func.value = step(
        job_name=DEFAULT_EVALUATION_JOB_NAME,
        name="predict",
        resources=resources,
        concurrency=concurrency,
        task_num=task_num,
        extra_kwargs=dict(
            ppl_batch_size=batch_size,
            ignore_error=not fail_on_error,
            ppl_auto_log=auto_log,
            ignore_dataset_data=not auto_log,
            dataset_uris=datasets,
        ),
    )(func)


def evaluate(*args: t.Any, **kw: t.Any) -> t.Any:
    """Defines an evaluate function.

    This function can be used as a decorator to define an evaluation-evaluate function that reduces the results of the
    predict function.

    Argument:
        use_predict_auto_log: [bool, optional] Passing the iterator of the predict auto-log results into the evaluate function.
            Default is True.
        resources: [Dict, optional] Resources for the predict task, such as memory, gpu etc. Current only supports
            the cloud instance.
    Examples:
    ```python
    from starwhale import evaluation

    @evaluation.evaluate
    def evaluate_results(predict_result_iter):
        ...

    @evaluation.evaluate(
        use_predict_auto_log=False
    )
    def evaluate_results():
        ...
    ```

    Returns:
        The decorated function.
    """
    if len(args) == 1 and len(kw) == 0 and callable(args[0]):
        return evaluate()(args[0])
    else:

        def _wrap(func: t.Callable) -> t.Any:
            _register_evaluate(func, **kw)
            setattr(func, DecoratorInjectAttr.Evaluate, True)
            return func

        return _wrap


_registered_evaluate_func = threading.local()


def _register_evaluate(
    func: t.Callable,
    resources: t.Optional[t.Dict[str, t.Any]] = None,
    use_predict_auto_log: bool = True,
) -> None:
    from .job import step

    try:
        val = _registered_evaluate_func.value
    except AttributeError:
        val = None

    if val is not None:
        raise RuntimeError("evaluate function can only be called once")

    _registered_evaluate_func.value = step(
        job_name=DEFAULT_EVALUATION_JOB_NAME,
        name="evaluate",
        resources=resources,
        concurrency=1,
        task_num=1,
        needs=["predict"],
        extra_kwargs=dict(
            ppl_auto_log=use_predict_auto_log,
        ),
    )(func)



from enum import Enum, unique

from starwhale.utils.error import NoSupportError


class InstanceType:
    STANDALONE = "standalone"
    CLOUD = "cloud"


class URIType:
    INSTANCE = "instance"
    PROJECT = "project"
    MODEL = "model"
    DATASET = "dataset"
    RUNTIME = "runtime"
    EVALUATION = "evaluation"
    UNKNOWN = "unknown"


class JobType:
    EVAL = "evaluation"


class EvalTaskType:
    ALL = "all"
    SINGLE = "single"


class RunSubDirType:
    DATASET = "dataset"
    STATUS = "status"
    RUNLOG = "runlog"
    LOG = "log"
    SWMP = "swmp"
    SWRT = "swrt"


class JobOperationType:
    CANCEL = "cancel"
    PAUSE = "pause"
    RESUME = "resume"
    CREATE = "create"
    REMOVE = "remove"
    RECOVER = "recover"
    INFO = "info"


class BundleType:
    MODEL = ".swmp"
    DATASET = ".swds"
    RUNTIME = ".swrt"


class RuntimeArtifactType:
    RUNTIME = "runtime_yaml"
    DEPEND = "dependencies"
    WHEELS = "wheels"
    FILES = "files"


class RuntimeLockFileType:
    VENV = "requirements-sw-lock.txt"
    CONDA = "conda-sw-lock.yaml"


def get_bundle_type_by_uri(uri_type: str) -> str:
    if uri_type == URIType.DATASET:
        return BundleType.DATASET
    elif uri_type == URIType.MODEL:
        return BundleType.MODEL
    elif uri_type == URIType.RUNTIME:
        return BundleType.RUNTIME
    else:
        raise NoSupportError(uri_type)


@unique
class DependencyType(Enum):
    PIP_PKG = "pip_pkg"
    PIP_REQ_FILE = "pip_req_file"
    CONDA_PKG = "conda_pkg"
    CONDA_ENV_FILE = "conda_env_file"
    WHEEL = "wheel"
    NATIVE_FILE = "native_file"



from __future__ import annotations

import os
import typing as t
from pathlib import Path

from rich.pretty import Pretty
from rich.syntax import Syntax

from starwhale.utils import console, load_yaml, pretty_bytes, in_production
from starwhale.consts import (
    PythonRunEnv,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    STANDALONE_INSTANCE,
)
from starwhale.base.uri import URI
from starwhale.base.type import URIType, InstanceType
from starwhale.base.view import BaseTermView
from starwhale.utils.venv import get_venv_env, get_conda_env, get_python_run_env
from starwhale.utils.error import NotFoundError, NoSupportError, ExclusiveArgsError
from starwhale.utils.config import SWCliConfigMixed

from .model import Runtime, RuntimeInfoFilter, StandaloneRuntime


class RuntimeTermView(BaseTermView):
    def __init__(self, runtime_uri: str | URI) -> None:
        super().__init__()

        if isinstance(runtime_uri, URI):
            self.uri = runtime_uri
        else:
            self.uri = URI(runtime_uri, expected_type=URIType.RUNTIME)

        self.runtime = Runtime.get_runtime(self.uri)

    @BaseTermView._simple_action_print
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.runtime.remove(force)

    @BaseTermView._simple_action_print
    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.runtime.recover(force)

    @BaseTermView._pager
    @BaseTermView._header
    def history(self, fullname: bool = False) -> t.List[t.Dict[str, t.Any]]:
        fullname = fullname or self.uri.instance_type == InstanceType.CLOUD
        return self._print_history(
            title="Runtime History", history=self.runtime.history(), fullname=fullname
        )

    def info(
        self,
        output_filter: RuntimeInfoFilter = RuntimeInfoFilter.basic,
    ) -> None:
        info = self.runtime.info()
        if not info:
            console.print(
                f":anguished_face: No runtime info found: {self.uri}", style="red"
            )
            return

        basic_content = Pretty(info["basic"], expand_all=True)
        runtime_content = Syntax(
            info.get("runtime_yaml", ""), "yaml", theme="ansi_dark"
        )
        manifest_content = Pretty(info.get("manifest", {}), expand_all=True)
        _locks = []
        for fname, content in info.get("lock", {}).items():
            _locks.append(f"#lock file: {fname}")
            _locks.append(content)
        lock_content = "\n".join(_locks)

        if output_filter == RuntimeInfoFilter.basic:
            console.print(basic_content)
        elif output_filter == RuntimeInfoFilter.runtime_yaml:
            console.print(runtime_content)
        elif output_filter == RuntimeInfoFilter.lock:
            console.print(lock_content)
        elif output_filter == RuntimeInfoFilter.manifest:
            console.print(manifest_content)
        else:
            console.rule("[green bold] Runtime Basic Info")
            console.print(basic_content)
            console.rule("[green bold] runtime.yaml")
            console.print(runtime_content)
            console.rule("[green bold] _manifest.yaml")
            console.print(manifest_content)
            console.rule("[green bold] Runtime Lock Files")
            console.print(lock_content)

    @classmethod
    @BaseTermView._only_standalone
    def activate(cls, uri: URI, force_restore: bool = False) -> None:
        Runtime.activate(uri, force_restore)

    @BaseTermView._only_standalone
    def dockerize(
        self,
        tags: t.List[str],
        push: bool,
        platforms: t.List[str],
        dry_run: bool,
        use_starwhale_builder: bool,
        reset_qemu_static: bool,
    ) -> None:
        self.runtime.dockerize(
            tags=tags,
            push=push,
            platforms=platforms,
            dry_run=dry_run,
            use_starwhale_builder=use_starwhale_builder,
            reset_qemu_static=reset_qemu_static,
        )

    @classmethod
    @BaseTermView._only_standalone
    def lock(
        cls,
        target_dir: str | Path,
        yaml_path: str | Path,
        env_name: str = "",
        env_prefix_path: str = "",
        no_cache: bool = False,
        stdout: bool = False,
        include_editable: bool = False,
        include_local_wheel: bool = False,
        emit_pip_options: bool = False,
        env_use_shell: bool = False,
    ) -> None:
        Runtime.lock(
            target_dir=target_dir,
            yaml_path=yaml_path,
            env_name=env_name,
            env_prefix_path=env_prefix_path,
            stdout=stdout,
            no_cache=no_cache,
            include_editable=include_editable,
            include_local_wheel=include_local_wheel,
            emit_pip_options=emit_pip_options,
            env_use_shell=env_use_shell,
        )

    @classmethod
    @BaseTermView._only_standalone
    def build_from_docker_image(
        cls, image: str, runtime_name: str = "", project: str = ""
    ) -> None:
        runtime_name = runtime_name or image.split(":")[0].split("/")[-1]
        runtime_uri = cls.prepare_build_bundle(
            project=project, bundle_name=runtime_name, typ=URIType.RUNTIME
        )

        rt = Runtime.get_runtime(runtime_uri)
        rt.build_from_docker_image(image=image, runtime_name=runtime_name)

    @classmethod
    @BaseTermView._only_standalone
    def build_from_python_env(
        cls,
        runtime_name: str = "",
        conda_name: str = "",
        conda_prefix: str = "",
        venv_prefix: str = "",
        project: str = "",
        cuda: str = "",
        cudnn: str = "",
        arch: str = "",
        download_all_deps: bool = False,
        include_editable: bool = False,
        include_local_wheel: bool = False,
    ) -> URI:
        set_args = list(filter(bool, (conda_name, conda_prefix, venv_prefix)))
        if len(set_args) >= 2:
            raise ExclusiveArgsError(
                f"conda_prefix({conda_prefix}), conda_name({conda_name}), venv_prefix({venv_prefix}) are the mutex args."
            )

        if conda_name:
            mode = PythonRunEnv.CONDA
            candidate_runtime_name = conda_name
        elif conda_prefix:
            mode = PythonRunEnv.CONDA
            candidate_runtime_name = conda_prefix
        elif venv_prefix:
            mode = PythonRunEnv.VENV
            candidate_runtime_name = venv_prefix
        else:
            mode = get_python_run_env()
            candidate_runtime_name = (
                get_conda_env() if mode == PythonRunEnv.CONDA else get_venv_env()
            ) or "default"

        if runtime_name == "":
            runtime_name = candidate_runtime_name.strip("/").split("/")[-1]

        runtime_uri = cls.prepare_build_bundle(
            project=project, bundle_name=runtime_name, typ=URIType.RUNTIME
        )
        rt = Runtime.get_runtime(runtime_uri)
        rt.build_from_python_env(
            runtime_name=runtime_name,
            mode=mode,
            conda_name=conda_name,
            conda_prefix=conda_prefix,
            venv_prefix=venv_prefix,
            cuda=cuda,
            cudnn=cudnn,
            arch=arch,
            download_all_deps=download_all_deps,
            include_editable=include_editable,
            include_local_wheel=include_local_wheel,
        )
        return runtime_uri

    @classmethod
    @BaseTermView._only_standalone
    def build_from_runtime_yaml(
        cls,
        workdir: str | Path,
        yaml_path: str | Path,
        runtime_name: str = "",
        project: str = "",
        download_all_deps: bool = False,
        include_editable: bool = False,
        include_local_wheel: bool = False,
        no_cache: bool = False,
        disable_env_lock: bool = False,
    ) -> URI:
        workdir = Path(workdir)
        yaml_path = Path(yaml_path)

        if not yaml_path.exists():
            raise NotFoundError(f"not found runtime yaml:{yaml_path}")
        _config = load_yaml(yaml_path)
        runtime_name = runtime_name or _config["name"]

        _runtime_uri = cls.prepare_build_bundle(
            project=project, bundle_name=runtime_name, typ=URIType.RUNTIME
        )
        if include_editable:
            console.print(
                ":bell: [red bold]runtime will include pypi editable package[/] :bell:"
            )
        else:
            console.print(
                ":bird: [red bold]runtime will ignore pypi editable package[/]"
            )

        _rt = Runtime.get_runtime(_runtime_uri)
        _rt.build_from_runtime_yaml(
            workdir=workdir,
            yaml_path=yaml_path,
            download_all_deps=download_all_deps,
            include_editable=include_editable,
            include_local_wheel=include_local_wheel,
            no_cache=no_cache,
            disable_env_lock=disable_env_lock,
        )
        return _runtime_uri

    @BaseTermView._only_standalone
    def extract(self, force: bool = False, target: t.Union[str, Path] = "") -> None:
        console.print(":oncoming_police_car: try to extract ...")
        path = self.runtime.extract(force, target)
        console.print(f":clap: extracted @ {path.resolve()} :tada:")

    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.List[str]] = None,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        filters = filters or []
        _uri = URI(project_uri, expected_type=URIType.PROJECT)
        cls.must_have_project(_uri)
        fullname = fullname or (_uri.instance_type == InstanceType.CLOUD)
        _runtimes, _pager = Runtime.list(_uri, page, size, filters)
        _data = BaseTermView.list_data(_runtimes, show_removed, fullname)
        return _data, _pager

    @classmethod
    @BaseTermView._only_standalone
    def quickstart_from_uri(
        cls,
        workdir: t.Union[Path, str],
        name: str,
        uri: URI,
        force: bool = False,
        disable_restore: bool = False,
    ) -> None:
        console.print(
            f":construction: quickstart Starwhale Runtime[{name}] environment from runtime URI({uri})..."
        )
        _sw_config = SWCliConfigMixed()
        if _sw_config.current_instance != STANDALONE_INSTANCE:
            raise NoSupportError(f"{_sw_config.current_instance} quickstart")
        StandaloneRuntime.quickstart_from_uri(
            workdir, name, uri, force, disable_restore
        )
        console.print(":clap: Starwhale Runtime environment is ready to use :tada:")

    @classmethod
    @BaseTermView._only_standalone
    def quickstart_from_ishell(
        cls,
        workdir: t.Union[Path, str],
        name: str,
        mode: str,
        disable_create_env: bool = False,
        force: bool = False,
        interactive: bool = False,
    ) -> None:
        console.print(
            f":construction: quickstart Starwhale Runtime[{name}] environment..."
        )
        StandaloneRuntime.quickstart_from_ishell(
            workdir, name, mode, disable_create_env, force, interactive
        )
        console.print(":clap: Starwhale Runtime environment is ready to use :tada:")

    @classmethod
    @BaseTermView._only_standalone
    def restore(cls, target: str) -> None:
        if in_production() or (os.path.exists(target) and os.path.isdir(target)):
            workdir = Path(target)
        else:
            _uri = URI(target, URIType.RUNTIME)
            _runtime = StandaloneRuntime(_uri)
            workdir = _runtime.store.snapshot_workdir
            if not workdir.exists():
                _runtime.extract(force=True, target=workdir)

        console.print(
            f":golfer: try to restore python runtime environment: {workdir} ..."
        )
        Runtime.restore(workdir)

    @classmethod
    def copy(
        cls,
        src_uri: str,
        dest_uri: str,
        force: bool = False,
        dest_local_project_uri: str = "",
    ) -> None:
        Runtime.copy(src_uri, dest_uri, force, dest_local_project_uri)
        console.print(":clap: copy done.")

    @BaseTermView._header
    def tag(
        self, tags: t.List[str], remove: bool = False, ignore_errors: bool = False
    ) -> None:
        # TODO: refactor model/runtime/dataset tag view-model
        if remove:
            console.print(f":golfer: remove tags [red]{tags}[/] @ {self.uri}...")
            self.runtime.remove_tags(tags, ignore_errors)
        else:
            console.print(f":surfer: add tags [red]{tags}[/] @ {self.uri}...")
            self.runtime.add_tags(tags, ignore_errors)


class RuntimeTermViewRich(RuntimeTermView):
    @classmethod
    @BaseTermView._pager
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.List[str]] = None,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        filters = filters or []
        _data, _pager = super().list(
            project_uri, fullname, show_removed, page, size, filters
        )

        custom_column: t.Dict[str, t.Callable[[t.Any], str]] = {
            "tags": lambda x: ",".join(x),
            "size": lambda x: pretty_bytes(x),
            "runtime": cls.place_holder_for_empty(),
        }

        cls.print_header()
        cls.print_table("Runtime List", _data, custom_column=custom_column)
        return _data, _pager


class RuntimeTermViewJson(RuntimeTermView):
    @classmethod
    def list(  # type: ignore
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.List[str]] = None,
    ) -> None:
        filters = filters or []
        _data, _pager = super().list(
            project_uri, fullname, show_removed, page, size, filters
        )
        cls.pretty_json(_data)

    def info(
        self,
        output_filter: RuntimeInfoFilter = RuntimeInfoFilter.basic,
    ) -> None:
        info = self.runtime.info()

        if output_filter == RuntimeInfoFilter.basic:
            info = {"basic": info.get("basic", {})}
        elif output_filter == RuntimeInfoFilter.lock:
            info = {"lock": info.get("lock", {})}
        elif output_filter == RuntimeInfoFilter.manifest:
            info = {"manifest": info.get("manifest", {})}
        elif output_filter == RuntimeInfoFilter.runtime_yaml:
            info = {"runtime_yaml": info.get("runtime_yaml", "")}

        self.pretty_json(info)

    def history(self, fullname: bool = False) -> None:
        fullname = fullname or self.uri.instance_type == InstanceType.CLOUD
        _data = BaseTermView.get_history_data(
            history=self.runtime.history(), fullname=fullname
        )
        self.pretty_json(_data)


def get_term_view(ctx_obj: t.Dict) -> t.Type[RuntimeTermView]:
    return (
        RuntimeTermViewJson if ctx_obj.get("output") == "json" else RuntimeTermViewRich
    )
